import numpy as np
from ..superconductivity import get_eh_sector
from ..superconductivity import build_nambu_matrix
from ..multihopping import MultiHopping
from .. import algebra

def extract_anomalous_dict(dd):
    """Given a dictionary, extract the anomalous part"""
    out = dict()
    for key in dd:
        d = dd[key] # get this patrix
        m01 = get_eh_sector(d,i=0,j=1)
        m10 = get_eh_sector(d,i=1,j=0)
        m = build_nambu_matrix(m01*0.0,c12=m01,c21=m10) # build matrix
        out[key] = m
    return out # return dictionary


def extract_normal_dict(dd):
    """Given a dictionary, extract the anomalous part"""
    out = dict()
    for key in dd:
        d = dd[key] # get this patrix
        m00 = get_eh_sector(d,i=0,j=0)
        m = build_nambu_matrix(m00) # build matrix
        out[key] = m
    return out # return dictionary




def get_anomalous_hamiltonian(self):
    """Return anomalous part of a Hamiltonian"""
    h0 = self.copy() ; h0.remove_nambu() ; h0.setup_nambu_spinor()
    h = self - h0
    return h


def get_singlet_hamiltonian(self):
    """Return only the anomalous spin singlet superconducting state"""
    self = self.copy()
    dd = self.get_multihopping().get_dict() # return the dictionary
    dd = extract_singlet_dict(dd)
    self.set_multihopping(MultiHopping(dd))
    return self


def get_triplet_hamiltonian(self):
    """Return only the anomalous spin singlet superconducting state"""
    h = get_anomalous_hamiltonian(self)
    return h - get_singlet_hamiltonian(h)



def extract_pairing(m):
  """Extract the pairing from a matrix, assuming it has the Nambu form"""
  nr = m.shape[0]//4 # number of positions
  uu = np.array(np.zeros((nr,nr),dtype=np.complex_)) # zero matrix
  dd = np.array(np.zeros((nr,nr),dtype=np.complex_)) # zero matrix
  ud = np.array(np.zeros((nr,nr),dtype=np.complex_)) # zero matrix
  for i in range(nr): # loop over positions
    for j in range(nr): # loop over positions
        ud[i,j] = m[4*i,4*j+2]
        dd[i,j] = m[4*i+1,4*j+2]
        uu[i,j] = m[4*i,4*j+3]
  return (uu,dd,ud) # return the three matrices



def extract_triplet_pairing(m):
  """Extract the pairing from a matrix, assuming it has the Nambu form"""
  m = algebra.todense(m) # dense matrix
  nr = m.shape[0]//4 # number of positions
  uu = np.array(np.zeros((nr,nr),dtype=np.complex_)) # zero matrix
  dd = np.array(np.zeros((nr,nr),dtype=np.complex_)) # zero matrix
  ud = np.array(np.zeros((nr,nr),dtype=np.complex_)) # zero matrix
  for i in range(nr): # loop over positions
    for j in range(nr): # loop over positions
        ud[i,j] = (m[4*i,4*j+2] - np.conjugate(m[4*j+3,4*i+1]))/2.
        dd[i,j] = m[4*i+1,4*j+2]
        uu[i,j] = m[4*i,4*j+3]
  return (uu,dd,ud) # return the three matrices


def extract_singlet_dict(dd):
    """Given a dictionary, extract the anomalous singlet part"""
    out = dict()
    for key in dd:
        d = dd[key] # get this matrix
        nr = d.shape[0]//4 # number of positions
        m = np.zeros(d.shape,dtype=np.complex_) # initialize
        m0 = dd[key]
        key2 = (-key[0],-key[1],-key[2]) 
        m1 = dd[key2]
        for i in range(nr): # loop over positions
            for j in range(nr): # loop over positions
                m[4*i,4*j+2] = (m0[4*i,4*j+2]+np.conjugate(m1[4*j+3,4*i+1]))/2.
                m[4*j+3,4*i+1] = (m0[4*j+3,4*i+1]+np.conjugate(m1[4*i,4*j+2]))/2.
                m[4*i+2,4*j] = (m0[4*i+2,4*j]+np.conjugate(m1[4*j+1,4*i+3]))/2.
                m[4*j+1,4*i+3] = (m0[4*j+1,4*i+3]+np.conjugate(m1[4*i+2,4*j]))/2.
        out[key] = m
    return out # return dictionary


def extract_triplet_dict(dd):
    dd = extract_anomalous_dict(dd) # anomalous part
    ds = extract_singlet_dict(dd) # singlet part
    from ..multihopping import MultiHopping
    return (MultiHopping(dd) - MultiHopping(ds)).get_dict()



def extract_singlet_pairing(m):
  """Extract the pairing from a matrix, assuming it has the Nambu form"""
  nr = m.shape[0]//4 # number of positions
  ud = np.array(np.zeros((nr,nr),dtype=np.complex_)) # zero matrix
  for i in range(nr): # loop over positions
    for j in range(nr): # loop over positions
        ud[i,j] = (m[4*i,4*j+2] + np.conjugate(m[4*j+3,4*i+1]))/2.
  return ud




def extract_singlet_hamiltonian(h):
    """Given a Hamiltonian, return only the superconducting singlet term"""
    h = h.copy() # copy Hamiltonian
    dd = h.get_dict() # extract dictionary
    dd = extract_singlet_dict(dd) # extract the singlet
    from ..multicell import set_dictionary
    h = set_dictionary(h,dd) # set the dictionary
    return h # return the singlet Hamiltonian


def extract_triplet_hamiltonian(h):
    """Given a Hamiltonian, return only the superconducting triplet term"""
    h = get_anomalous_hamiltonian(h)
    return h - extract_singlet_hamiltonian(h) # return the triplet



def extract_custom_pairing(m,mode="all"):
    """Given a matrix, extract the pairing matrix according to some rule"""
    raise # this may be buggy
    if mode=="singlet": # singlet, with sign
        m = extract_singlet_pairing(m) # matrix with pairings 
        return m
    elif mode=="triplet": # triplet, summed over
        ms = extract_triplet_pairing(m) # matrix with pairings 
        m = np.sum(np.abs(np.array(ms))**2,axis=0)
        return m
    elif mode=="all": # compute all in absolute value
        ms = extract_triplet_pairing(m) # matrix with pairings (3 of them)
        mt = np.sum(np.abs(np.array(ms))**2,axis=0)
        m = extract_singlet_pairing(m) # matrix with pairings 
        m = mt + np.abs(np.array(m))**2 # singlet plus triplet
        return m
    elif mode=="both": # singlet and triplet with interference effects
        if m.shape[0]==4: # this is a quick fix for single site models
            m = m@np.conjugate(m.T)
            return np.array([[np.trace(m)]])
        else: raise
    else: raise



def extract_pairing_kmap(h,write=False,i=None,j=None,mode="all",**kwargs):
    """Extract the pairing in reciprocal space"""
    if not h.has_eh: raise # not implemented
    h = get_anomalous_hamiltonian(h)
    if j is None: j = i # same site is the default
    if mode=="all": pass # do nothing 
    elif mode=="singlet": h = extract_singlet_hamiltonian(h) # singlet
    elif mode=="triplet": h = extract_triplet_hamiltonian(h) # triplet
    else: raise
    fk = h.get_hk_gen() # Bloch Hamiltonian generator
    def f0(k):
        m = fk(k) # full k-dependent Hamiltonian
        if i is None: return np.trace(m@np.conjugate(m.T))/m.shape[0]
        else: raise # not implemented
        # return m[i,j] # return pairing
    from .. import spectrum
    (ks,ds) = spectrum.reciprocal_map(h,f0,write=write,**kwargs)
    return ks[:,0],ks[:,1],ds
#    dref = f0(np.random.random(3)) ; dref = dref/np.abs(dref) # reference
#    fr = lambda k: (f0(k)/dref).real # reference
#    fi = lambda k: (f0(k)/dref).imag # reference
#    from .. import spectrum
#    (ks,dsr) = spectrum.reciprocal_map(h,fr,write=write,**kwargs)
#    (ks,dsi) = spectrum.reciprocal_map(h,fi,write=write,**kwargs)
#    return ks[:,0],ks[:,1],dsr+1j*dsi


def extract_absolute_pairing(h,mode="singlet",**kwargs):
    """Extract the absolute value of the SC order in the BZ"""
    h = get_anomalous_hamiltonian(h)
    if mode=="all" or mode=="both": pass # do nothing 
    elif mode=="singlet": h = extract_singlet_hamiltonian(h) # singlet
    elif mode=="triplet": h = extract_triplet_hamiltonian(h) # triplet
    else: raise # do nothing
    fk = h.get_hk_gen() # Bloch Hamiltonian generator
    from ..klist import kmesh
    ks = kmesh(h.dimensionality,**kwargs) # kpoints
    def f(k):
        m = fk(k) # return Bloch Hamiltonian
        return np.trace(m@m)/m.shape[0]
    return np.sqrt(np.mean([f(k) for k in ks])) # return mean value




def extract_absolute_spatial_pairing(h,mode="singlet",**kwargs):
    """Extract the absolute value of the SC order in the BZ"""
    h = get_anomalous_hamiltonian(h) # overwrite Hamiltonian
    if mode=="all" or mode=="both": pass # do nothing
    elif mode=="singlet": h = extract_singlet_hamiltonian(h) # singlet
    elif mode=="triplet": h = extract_triplet_hamiltonian(h) # triplet
    else: raise # do nothing
    fk = h.get_hk_gen() # Bloch Hamiltonian generator
    from ..klist import kmesh
    ks = kmesh(h.dimensionality,**kwargs) # kpoints
    def f(k):
        m = fk(k) # return Bloch Hamiltonian
        m2 = m@m # compute H**2
        out = np.diag(m2) # return the diagonal
        return out
    out = np.mean([f(k) for k in ks],axis=0) # return mean value
    from ..increase_hilbert import full2profile
    out = full2profile(h,out) # resum
    return np.sqrt(out.real/2.) # return the spatial profile






import numpy as np
import scipy.linalg as lg
from . import parallel
from numba import jit
from . import algebra


try:
    from . import chif90
    use_fortran = True
except:
    use_fortran = False
#    print("Error, chif90 not found")

use_fortran = False


def chargechi(h,i=0,j=0,es=np.linspace(-3.0,3.0,100),delta=0.01,temp=1e-7):
    """Compute charge response function"""
    if h.dimensionality!=0: raise
    hk = h.get_hk_gen() # get generator
    m = hk(0) # get Hamiltonian
    esh,ws = algebra.eigh(m)
    ws = np.transpose(ws)
    if i<0: raise
    if j<0: raise
#    if use_fortran:
#      return es,chif90.elementchi(ws,esh,ws,esh,es,i+1,j+1,temp,delta)
#    else:
    out = 0*es + 0j # initialize
    return es,elementchi(ws,esh,ws,esh,es,i,j,temp,delta,out)

@jit(nopython=True)
def elementchi(ws1,es1,ws2,es2,omegas,ii,jj,T,delta,out):
    """Compute the response function"""
    out  = out*0.0 # initialize
    n = len(ws1) # number of wavefunctions
    for i in range(n): # first loop over states
      oi = es1[i]<0.0 # first occupation
      for j in range(n): # second loop over states
          oj = es2[j]<0.0 # second occupation
          fac = ws1[i][ii]*ws2[j][ii] # add the factor
          fac *= np.conjugate(ws1[i][jj]*ws2[j][jj]) # add the factor
          fac *= oi - oj # occupation factor
          out = out + fac*(1./(es1[i]-es2[j] - omegas + 1j*delta))
    return out

def chargechi_nowf(h,i=0,j=0,es=np.linspace(-3.0,3.0,100),delta=0.01,temp=1e-7):
    """Compute charge response function"""
    if h.dimensionality!=0: raise
    hk = h.get_hk_gen() # get generator
    m = hk(0) # get Hamiltonian
    esh,ws = algebra.eigh(m)
    ws = np.transpose(ws)*0. + 1.
    if i<0: raise
    if j<0: raise
    return es,chif90.elementchi(ws,esh,ws,esh,es,i+1,j+1,temp,delta)







def chargechi_row(h,i=0,es=np.linspace(-3.0,3.0,100),delta=1e-6,temp=1e-7):
    """Compute charge response function"""
    if h.dimensionality!=0: raise
    hk = h.get_hk_gen() # get generator
    m = hk(0) # get Hamiltonian
    esh,ws = algebra.eigh(m)
    ws = np.transpose(ws)
    out = [] # store
    if i<0: raise
    f = lambda j: elementchi(ws,esh,ws,esh,es,i+1,j+1,temp,delta)
    out = parallel.pcall(f,range(m.shape[0])) # parallel call
    return np.array(out)



def chargechi_nowf(h,i=0,j=0,es=np.linspace(-3.0,3.0,100),delta=0.01,temp=1e-7):
    """Compute charge response function"""
    if h.dimensionality!=0: raise
    hk = h.get_hk_gen() # get generator
    m = hk(0) # get Hamiltonian
    esh,ws = algebra.eigh(m)
    ws = np.transpose(ws)*0. + 1.
    if i<0: raise
    if j<0: raise
    return es,chif90.elementchi(ws,esh,ws,esh,es,i+1,j+1,temp,delta)



def chargechi_reciprocal(h,i=None,
        es=np.linspace(-4.,4.,200),delta=1e-3,**kwargs):
    """Return the charge susceptibility in reciprocal space"""
    if h.dimensionality!=0: raise
    if i is None: i = h.geometry.get_central(1)[0]
    # compute correlators
    cs = chargechi_row(h,es=es,delta=delta,**kwargs) 
    # get the distances
    dx = np.abs(h.geometry.r[:,0] - h.geometry.r[i,0])
    # now do the interpolation
    from scipy.interpolate import interp1d
    xs = np.linspace(0.,np.max(dx),len(dx))
    cs = np.array([interp1d(xs,ic)(xs) for ic in cs.T]).T
#    if len(cs)!=len(h.geometry.r[:,0]): raise
    freq = np.fft.fftfreq(len(cs[:,0])) # get the frequencies
    # perform the fourier transform
    fcs = np.array([np.fft.fft(ic) for ic in cs.T])
    fo = open("CHIK.OUT","w")
    print(fcs.shape,es.shape,freq.shape)
    for i in range(len(freq)):
      for j in range(len(es)):
#          fo.write(str(h.geometry.r[i,0])+"  ")
          fo.write(str(freq[i])+"  ")
          fo.write(str(es[j])+"  ")
          fo.write(str(np.abs(fcs[j][i]))+"\n")
    fo.close()



def chiAB(h,energies=np.linspace(-3.0,3.0,100),q=[0.,0.,0.],nk=60,
               delta=0.1,temp=1e-7,A=None,B=None,projs=None,
               mode="matrix"):
    """Compute AB response function
       - energies: energies of the dynamical response
       - q: q-vector of the response
       - nk: number of k-points for the integration
       - delta: imaginary part
       - A: first operator
       - B: second operator
       - projs: projection operators
       - mode: response to compute"""
    hk = h.get_hk_gen() # get generator
    if A is None or B is None:
        A = np.identity(h.intra.shape[0],dtype=np.complex_)
        B = A # initial operator
    # generate the projectors
    if projs is None:
        from . import operators
        projs = [operators.index(h,n=[i]) for i in range(len(h.geometry.r))]
    def getk(k):
        m1 = hk(k) # get Hamiltonian
        es1,ws1 = algebra.eigh(m1)
        ws1 = np.array(ws1.T,dtype=np.complex_)
        m2 = hk(k+q) # get Hamiltonian
        es2,ws2 = algebra.eigh(m2)
        ws2 = np.array(ws2.T,dtype=np.complex_)
        def getAB(Ai,Bj): # compute for a single operator
            out = 0*energies + 0j # initialize
            return chiAB_jit(ws1,es1,ws2,es2,energies,Ai,Bj,temp,delta,out)
        if mode=="matrix": # return a matrix
            out = np.array([[getAB(pi@A,pj@B) for pi in projs] for pj in projs])
            return np.transpose(out,(2,0,1)) # return array of matrices
        elif mode=="trace": # return the trace
            out = np.array([getAB(pi@A,pi@B) for pi in projs])
            return np.mean(out,axis=0) # sum over the first axis
        else: raise # not implemented
    ks = h.geometry.get_kmesh(nk=nk) # get the kmesh
    out = np.mean([getk(k) for k in ks],axis=0) # sum over kpoints
    return out


def chiAB_trace(h,**kwargs):
    """Compute the trace of chiAB"""
    return chiAB(h,mode="trace",**kwargs)




def chiABmap(h,energies=np.linspace(-3.0,3.0,100),nq=30,
                qpath=None,**kwargs):
    """Return the map for the response function"""
    if qpath is None: qpath = h.geometry.get_default_kpath(nk=nq)
    return None



@jit(nopython=True)
def chiAB_jit(ws1,es1,ws2,es2,omegas,A,B,T,delta,out):
    """Compute the response function"""
    out  = out*0.0 # initialize
    n = len(ws1) # number of wavefunctions
    for i in range(n):
      if es1[i]<0.0: oi = 1.0 # first occupation
      else: oi = 0.0
      for j in range(n):
          if es2[j]<0.0: oj = 1.0 # second occupation
          else: oj = 0.0
          fac = np.conjugate(ws1[i]).dot(A@ws2[j]) # add the factor
          fac *= np.conjugate(ws2[j]).dot(B@ws1[i]) # add the factor
          fac *= oi - oj # occupation factor
          out = out + fac*(1./(es1[i]-es2[j] - omegas + 1j*delta))
    return out



from .chitk.static import chargechi as static_charge_correlator
from .chitk.static import szchi as static_sz_correlator
from .chitk.static import sxchi as static_sx_correlator
from .chitk.static import sychi as static_sy_correlator






from __future__ import print_function
import numpy as np
from . import geometry
from . import sculpt



def get_geometry(shape="polygon",**kwargs):
    if shape=="polygon":
        return get_polygon_island(**kwargs)
    elif shape=="flower":
        return get_flower_island(**kwargs)
    else: raise



def default_geometry_edges(name="square",nedges=None,geo=None):
    lattice_name = name
    if lattice_name=="honeycomb":
      geometry_builder = geometry.honeycomb_lattice
      if nedges==None: nedges = 6
    elif lattice_name=="square":
      geometry_builder = geometry.square_lattice
      if nedges==None: nedges = 4
    elif lattice_name=="kagome":
      geometry_builder = geometry.kagome_lattice
      if nedges==None: nedges = 3
    elif lattice_name=="lieb":
      geometry_builder = geometry.lieb_lattice
      if nedges==None: nedges = 4
    elif lattice_name=="triangular":
      geometry_builder = geometry.triangular_lattice
      if nedges==None: nedges = 3
    else: raise
    # first create a raw unit cell
    if geo is not None: 
      print("Geometry generator taken from input")
      g = geo # builder is input geometry
    else:
      g = geometry_builder()  # build a 2d unit cell
    return g,nedges # get the geometry



def get_polygon_island(n=5,rot=0.,clean=True,shift=[0.,0.,0.],**kwargs):
    """ Create a 0d polygon island"""
    g,nedges = default_geometry_edges(**kwargs)
    nf = float(n)   # get the desired size, in float
    g = g.supercell(int(7*n))   # create supercell
    g.set_finite() # set as finite system
    g.center() # center the geometry
    # now scuplt the geometry
    g = sculpt.rotate(g,rot) # initial rotation
    # now shift the lattice
    g.x += shift[0]
    g.y += shift[1]
    g.z += shift[2]
    g.xyz2r()
    def polygon(g):
        def f(r): 
            return r[0]>-nf*(np.cos(np.pi/3)+1.)  # function to use as cut
        for i in range(nedges): # loop over rotations, 60 degrees
            g = sculpt.intersec(g,f) # retain certain atoms
            g = sculpt.rotate(g,2.*np.pi/nedges) # rotate 60 degrees
        return g # return geometry
    g = polygon(g) # cut according to the polygon rule
    if clean: # if it is cleaned
      g = sculpt.remove_unibonded(g,iterative=True)  # remove single bonded
    g.center() # center the geometry
    return g



def get_flower_island(r=5.,dr=1.,clean=True,**kwargs):
    """ Create a 0d polygon island"""
    g,nedges = default_geometry_edges(**kwargs)
    nf = float(r)   # get the desired size, in float
    g = g.supercell(int(5*r)) # create supercell
    g.set_finite() # set as finite system
    g.set_origin()
    # now scuplt the geometry
    def flower(ri):
        phi = np.arctan2(ri[1],ri[0]) # angle
        rmax = r + dr*np.cos(nedges*phi) # maximum radius
        ri2 = np.sqrt(ri.dot(ri)) # location
        return ri2<rmax # if it is inside
    g = sculpt.intersec(g,flower) # retain certain atoms
    if clean: # if it is cleaned
      g = sculpt.remove_unibonded(g,iterative=True)  # remove single bonded
    g.center() # center the geometry
    return g






# library to create operators
from __future__ import division
import numpy as np
from scipy.sparse import csc_matrix as csc
from scipy.sparse import csc_matrix
from scipy.sparse import bmat,diags
from scipy.sparse import identity
from .superconductivity import build_eh
from scipy.sparse import issparse
import scipy.linalg as lg
from . import current
from . import algebra
from . import topology
from . import superconductivity
from .algebra import braket_wAw

import numbers

isnumber = algebra.isnumber


class Operator():
    def __init__(self,m,linear=True):
        """Initialization"""
        from .hamiltonians import Hamiltonian
        self.linear = linear
        self.matrix = None
        if algebra.ismatrix(m):
            self.m = lambda v,k=None: m@v # create dummy function
            self.matrix = m
            self.linear = True
        elif type(m)==Operator: 
            self.m = m.m
            self.linear = m.linear
            self.matrix = m.matrix
        elif isinstance(m, numbers.Number): 
            self.m = lambda v,k=None: m*v
        elif callable(m): 
        #    self.m = lambda v,k=None: m(v)
            self.m = m # as function
        elif type(m)==Hamiltonian: # Hamiltonian type 
            hkgen = m.get_hk_gen()
            self.m = lambda v,k=None: hkgen(k)@v
            self.linear = True
        else: 
            print("Unrecognised type",type(m))
            raise
    def __mul__(self,a):
        """Define the multiply method"""
        if type(a)==Operator:
            out = Operator(self)
            if self.matrix is not None and a.matrix is not None:
                out.matrix = self.matrix@a.matrix
                out.m = lambda v,k=None: out.matrix@v # create dummy function
            else: out.m = lambda v,k=None: self.m(a.m(v,k=k),k=k)
            out.linear = self.linear and a.linear
            return out
        elif algebra.ismatrix(a): # matrix type
            if self.matrix is not None: # return a matrix
                return self.matrix@a # multiply matrices
            else:
                return self*Operator(a) # convert to operator
        elif algebra.isvector(a): # array type
            return self(a)
        else:
            return self*Operator(a) # convert to operator
    def __rmul__(self,a):
        return Operator(a)*self
    def __truediv__(self,a):
        if isnumber(a): return self*(1./a)
        else: raise
    def __add__(self,a):
        """Define the add method"""
        if type(a)==Operator:
            out = Operator(self)
            out.m = lambda v,k=None: self.m(v,k=k) + a.m(v,k=k)
            if self.matrix is not None and a.matrix is not None:
                out.matrix = self.matrix + a.matrix
            out.linear = self.linear and a.linear
            return out
        else:
            return self + Operator(a) # convert to operator
    def __sub__(self,a):
        """Substraction method"""
        return self + (-a)
    def __neg__(self):
        """Negative operator"""
        out = Operator(self)
        out.m = lambda v,k=None: -self.m(v,k=k)
        return out
    def __call__(self,v,k=None):
        """Define the call method"""
        return self.m(v,k=k) 
    def __matmul__(self,a): return self*a
    def get_matrix(self,k=None):
        """Return matrix if possible"""
        if self.matrix is not None: return self.matrix
    def inv(self):
        """Return the inverse operator"""
        if self.matrix is not None and self.linear: # input is a matrix
            m = self.matrix
            def f(v,**kwargs):
                return algebra.applyinverse(m,v)
            return Operator(f,linear=True)
        else: raise # not implemented
    def braket(self,w,**kwargs):
        """Compute an expectation value"""
        wi = self(w,**kwargs) # apply the operator
        out =  algebra.braket_ww(w,wi)
        if np.abs(out.imag)<1e-8: return out.real
        else: return out


def object2operator(a):
    if a is None: return None
    else: return Operator(a)



def index(h,n=[0]):
  """Return a projector onto a site"""
  num = len(h.geometry.r)
  val = [1. for i in n]
  m = csc((val,(n,n)),shape=(num,num),dtype=np.complex_)
  m = h.spinless2full(m) # return matrix with e-h
  return Operator(m@m)



def rfunction2operator(h,f):
    """Given a function that takes a position, return the operator"""
    n = len(h.geometry.r)
    val = [f(ri) for ri in h.geometry.r]
    inds = range(n)
    m = csc((val,(inds,inds)),shape=(n,n),dtype=np.complex_)
    return h.spinless2full(m) # return matrix


def density2operator(h,d):
    """Given a function that takes a position, return the operator"""
    n = len(h.geometry.r)
    if len(d)!=n: raise
    inds = range(n)
    m = csc((d,(inds,inds)),shape=(n,n),dtype=np.complex_)
    return h.spinless2full(m) # return matrix




def operator2list(operator):
  """Convert an input operator in a list of operators"""
  if operator is None: # no operator given on input
    operator = [] # empty list
  elif not isinstance(operator,list): # if it is not a list
    operator = [operator] # convert to list
  return operator



def interface1d(h,cut = 3.):
  dind = 1 # index to which divide the positions
  if h.has_spin:  dind *= 2 # duplicate for spin
  if h.has_eh:  dind *= 2  # duplicate for eh
  n = h.intra.shape[0] # number of elments of the hamiltonian
  data = [] # epmty list
  for i in range(n): # loop over elements
    y = h.geometry.y[i//dind]
    if np.abs(y)<cut: data.append(1.) # if it belongs to the interface
    else:  data.append(0.)  # otherwise
  row, col = range(n),range(n)
  m = csc((data,(row,col)),shape=(n,n),dtype=np.complex_)
  return m # return the operator



def get_interface(h,fun=None):
  """Return an operator that projects onte the interface"""
  dind = 1 # index to which divide the positions
  if h.has_spin:  dind *= 2 # duplicate for spin
  if h.has_eh:  dind *= 2  # duplicate for eh
  iden = csc(np.matrix(np.identity(dind,dtype=np.complex_))) # identity matrix
  r = h.geometry.r # positions
  out = [[None for ri in r] for rj in r] # initialize
  if fun is None: # no input function
    cut = 2.0 # cutoff
    if h.dimensionality==1: index = 1
    elif h.dimensionality==2: index = 2
    else: raise
    def fun(ri): # define the function
      if np.abs(ri[index])<cut: return 1.0
      else: return 0.0
  for i in range(len(r)): # loop over positions
    out[i][i] = fun(r[i])*iden 
  return bmat(out) # return matrix



def get_pairing(h,ptype="s"):
  """Return an operator that calculates the expectation value of the
  s-wave pairing"""
  if not h.has_eh: raise # only for e-h systems
  if ptype=="s": op = superconductivity.spair
  elif ptype=="deltax": op = superconductivity.deltax
  elif ptype=="deltay": op = superconductivity.deltay
  elif ptype=="deltaz": op = superconductivity.deltaz
  else: raise
  r = h.geometry.r
  out = [[None for ri in r] for rj in r]
  for i in range(len(r)): # loop over positions
    out[i][i] = op
  return bmat(out) # return matrix



def get_electron(h):
  """Operator to project on the electron sector"""
  if not h.has_eh:
      return np.identity(h.intra.shape[0])
  elif h.check_mode("spinful_nambu"): # only for e-h systems
      op = superconductivity.proje
      r = h.geometry.r
      out = [[None for ri in r] for rj in r]
      for i in range(len(r)): # loop over positions
        out[i][i] = op
      return bmat(out)
  elif h.check_mode("spinless_nambu"):
      from .sctk import spinless
      return spinless.proje(h.intra.shape[0])
  else: raise


def get_hole(h):
  """Operator to project on the electron sector"""
  if not h.has_eh: raise # only for e-h systems
  op = superconductivity.projh
  r = h.geometry.r
  out = [[None for ri in r] for rj in r]
  for i in range(len(r)): # loop over positions
    out[i][i] = op
  return bmat(out)


def get_bulk(h,fac=0.8):
    """Return the bulk operator"""
    r = h.geometry.r # positions
    g = h.geometry
    g.center() # center the geometry
    out = np.array([1. for ir in r]) # initialize
    if h.dimensionality==0:
        dr = r[:,0]**2 + r[:,1]**2 # radii
        dr = dr - np.min(dr)
        dr = dr/np.max(dr) # to interval 0,1
        out[fac<dr] = 0.0 # set to zero
    elif h.dimensionality==1:
        dr = r[:,1] # y positions
        dr = dr - np.min(dr)
        dr = dr/np.max(dr) # to interval 0,1
        dr2 = dr - np.mean(dr) # minus the average
        out[fac/2.<np.abs(dr2)] = 0.0 # set to zero
    elif h.dimensionality==2:
        dr = r[:,2] # z positions
        dr = dr - np.min(dr)
        dr = dr/np.max(dr) # to interval 0,1
        dr2 = dr - np.mean(dr) # minus the average
        out[fac/2.<np.abs(dr2)] = 0.0 # set to zero
    else: return NotImplemented
    from scipy.sparse import diags
    n = len(r) # number of sites
    out = diags([out],offsets=[0],shape=(n,n),dtype=np.complex_) # create matrix
    m = h.spinless2full(out) # return this matrix
    return m@m # return the square


def get_surface(self,**kwargs):
    m = get_bulk(self,**kwargs)
    return identity(m.shape[0]) - m 


def bulk1d(h,p = 0.5):
    return get_bulk(h,fac=1.-p/2.)

def get_xposition(h):  return get_position(h,mode="x")
def get_yposition(h):  return get_position(h,mode="y")
def get_zposition(h):  return get_position(h,mode="z")




def get_position(h,mode="z"):
  dind = 1
  if h.has_spin:  dind *= 2 # duplicate for spin
  if h.has_eh:  dind *= 2  # duplicate for eh
  n = h.intra.shape[0] # number of elments of the hamiltonian
  if len(h.geometry.z)!=n//dind: raise # dimensions do not match
  data = [] # epmty list
  if mode=="x": pos = h.geometry.x
  elif mode=="y": pos = h.geometry.y
  elif mode=="z":  pos = h.geometry.z
  else: raise
  for i in range(n): # loop over elements
    z = pos[i//dind]
    data.append(z)
  row, col = range(n),range(n)
  m = csc((data,(row,col)),shape=(n,n),dtype=np.complex_)
  return m # return the operator



from .spin import sx,sy,sz # import pauli matrices
 

from .operatortk.spin import get_sx
from .operatortk.spin import get_sy
from .operatortk.spin import get_sz






def get_z(h):
  """Operator for the calculation of z expectation value"""
  if h.intra.shape[0]==len(h.geometry.z): # if as many positions as entries
    op = np.zeros(h.intra.shape,dtype=np.complex_) # initialize matrix
    for i in range(len(h.geometry.z)):
      op[i,i] = h.geometry.z[i]
    return op
  raise
  if h.has_eh: raise
  if not h.has_spin: raise
  if h.has_spin:
    op = np.zeros(h.intra.shape,dtype=np.complex_) # initialize matrix
    for i in range(len(op)//2):   
      op[2*i,2*i+1] = -1j
      op[2*i+1,2*i] = 1j
  



def get_rop(h,fun):
  """Operator for the calculation of a position expectation value"""
  rep = 1 # repetitions 
  if h.has_spin: rep *= 2
  if h.has_eh: rep *= 2
  data = []
  for ri in h.geometry.r: 
    for i in range(rep): data.append(fun(ri)) # store
  n = h.intra.shape[0]
  row = range(n)
  col = range(n)
  m = csc((data,(row,col)),shape=(n,n),dtype=np.complex_)
  return m




def get_sublattice(h,mode="both"):
  """Sublattice operator"""
  if not h.geometry.has_sublattice: raise
  rep = 1 # repetitions 
  if h.has_spin: rep *= 2
  if h.has_eh: rep *= 2
  data = []
  for s in h.geometry.sublattice: 
    for i in range(rep): 
      if mode=="both": data.append(s) # store
      elif mode=="A": data.append((s+1.)/2.) # store
      elif mode=="B": data.append((-s+1.)/2.) # store
      else: raise
  n = h.intra.shape[0]
  row = range(n)
  col = range(n)
  m = csc((data,(row,col)),shape=(n,n),dtype=np.complex_)
  return m


def get_velocity(h):
  """Return the velocity operator"""
  if h.dimensionality==1:
    vk = current.current_operator(h)
    def f(w,k=[0.,0.,0.]):
        return vk(k)@w
    return f
  elif h.dimensionality==2:
    def f(w,k=[0.,0.,0.]):
      vx = current.derivative(h,k,order=[0,1])
      vy = current.derivative(h,k,order=[1,0])
      R = np.array(h.geometry.get_k2K())
#      R = algebra.inv(R) # not sure if this is ok
      v = [braket_wAw(w,vx),braket_wAw(w,vy),0]
      v = np.array(v).real
      return (v@R@v)*w # return the scalar product
    return Operator(f)
  else: raise



get_current = get_velocity

def get_spin_current(h):
  vk = current.current_operator(h)
  sz = get_sz(h)
  def f(w,k=[0.,0.,0.]):
    return braket_wAw(w,vk(k)).real*braket_wAw(w,sz).real
  return f



from .operatortk.valley import get_valley



def get_inplane_valley(h):
  """Returns an operator that computes the absolute value
  of the intervalley mixing"""
  ho = h.copy() # copy Hamiltonian
  ho.clean() # set to zero
  ho.add_modified_haldane(1.0/4.5) # add modified Haldane coupling
  hkgen = ho.get_hk_gen() # get generator for the hk Hamiltonian
  hkgen0 = h.get_hk_gen() # get generator for the hk Hamiltonian
  def fun(w,k=None):
#    return abs(np.sum(w*w))
    if h.dimensionality>0 and k is None: raise # requires a kpoint
    hk = hkgen(k) # evaluate Hamiltonian
    hk0 = hkgen0(k) # evaluate Hamiltonian
    A = hk@hk0 - hk0@hk # commutator
    A = -A@A
    return abs(braket_wAw(w,A)) # return the braket
  return fun # return function





def tofunction(A):
    """Transform this object into a callable function"""
    if A is None: return lambda x,k=0.0: 1.0
    return Operator(A) # use operator
#    if A is None: return lambda x,k=0.0: 1.0 # no input
#    if callable(A): return A # if it is a function
#    else: return lambda x,k=0.0: braket_wAw(x,A).real # if it is a matrix


def ipr(w,k=None):
    """IPR operator"""
    return np.sum(np.abs(w)**4)*w # return a vector


def get_envelop(h,sites=[],d=0.3):
    """
    Return a list of operators that project on the different
    sites
    """
    # get a first neighbor Hamiltonian
    h0 = h.geometry.get_hamiltonian(has_spin=h.has_spin,is_sparse=True)
    m = h0.get_hk_gen()([0.,0.,0.]) # evaluate Hamiltonian at Gamma
    out = [] # output list
    for s in sites: # loop over sites
      c = m.getcol(s) # get column
      c = np.array(c.todense()) # transform into a dense matrix
      c = c.reshape(m.shape[0]) # 1D vector
      c = c*d # renormalize all the hoppings
      c[s] = 1.0 # set same atom to 1
      c = c/np.sum(c) # normalize the whole vector
      c = diags([c],[0],dtype=np.complex_) # create matrix
      out.append(c) # store matrix
    return out # return matrices


def get_sigma_minus(h):
    """
    Return the sublattice Pauli matrix \sigma_-
    """
    def fun(r1,r2):
        i1 = h.geometry.get_index(r1,replicas=True)
        if not h.geometry.sublattice[i1]==1: return 0.0
        i2 = h.geometry.get_index(r2,replicas=True)
        dr = r1-r2 # distance
        if 0.9<dr.dot(dr)<1.1: return 1.0 # get first neighbor
        return 0.0
    h0 = h.geometry.get_hamiltonian(has_spin=h.has_spin,fun=fun) # FN coupling
    hk = h0.get_hk_gen() # get generator
    return hk # return function





def get_valley_taux(h):
    """
    Return the tau x valley operator
    """
    g = h.geometry
    h0 = g.get_hamiltonian(has_spin=False) # FN coupling
    h0.turn_multicell() # multicell Hamiltonian
    h1 = h0.copy() # new Hmailtonian
    h1.clean()
    hs = h1.copy() # for sublattice
    hs.add_onsite(0.5) ; hs.add_sublattice_imbalance(0.5) 
    h1.add_kekule(1.0) # add kekule term 
    hops = hs.get_multihopping() # Multihopping for sublattice
    hop0 = h0.get_multihopping() # MultiHopping object
    hop1 = h1.get_multihopping() # MultiHopping object
    # kill hoppings that start in the A sublattice
    hop0 = hop0*hops
   # hop1 = hop0*hops
    # now define the valley mixing
    hop2 = 1j*hop0*hop1 
    hop2 = hop1
    hop2 = hop2 + hop2.get_dagger() # make it Hermitian
    hop2 = hop2.get_dict() # get the dictionary
    h2 = h0.copy() # dummy Hamiltonian
    h2.clean()
    h2.intra = hop2[(0,0,0)] # onsite matrix
    del hop2[(0,0,0)] # onsite matrix
    h2.hoppings = hop2 # store the rest of the hoppings
    hk = h2.get_hk_gen() # get the generating function
    return Operator(lambda m=None,k=None: m@hk(k)) # return operator








def get_operator(op,k=[0.,0.,0.],h=None):
    """Get a function that acts as an operator"""
    return Operator(op)


def get_berry(h,**kwargs):
    """Return Berry operator"""
    return topology.berry_operator(h,**kwargs)

def get_valley_berry(h,**kwargs):
    """Return Valley Berry operator"""
    return get_operator_berry(h,"valley",**kwargs)


def get_operator_berry(h,name,**kwargs):
    """Return Valley Berry operator"""
    op = h.get_operator(name,return_matrix=True)
    return topology.berry_operator(h,operator=op,**kwargs)



def get_sz_berry(h,**kwargs):
    """Return Valley Berry operator"""
    return get_operator_berry(h,"sz",**kwargs)


def get_matrix_operator(h,name,k=None,**kwargs):
    """Return a function that takes a matrix as input and returns another
    matrix"""
    if name=="valley":
        op = get_valley(h,projector=True) # valley operator
        return op
    elif name in ["valley_spin","spin_valley","valley_sz","sz_valley"]:
        op = get_valley(h,projector=True) # valley operator
        sz = h.get_operator("sz")
        return lambda m,k=None: op(m,k=k)@sz # return operator
    else:
        op = h.get_operator(name) # assume that it is a matrix
        return lambda m,k=None: op@m


def bool_layer_array(g,n=0):
    """Return the lowest layer array"""
    fac = []
    z0 = sorted(np.unique(g.z).tolist())[n]
    fac = g.z*0. # initialize
    fac[np.abs(g.z-z0)<1e-3] = 1.0
#    for z in g.z:
#        if abs(z-z0)<1e-3: fac.append(1)
#        else: fac.append(0)
#    fac = np.array(fac)
    return fac


bottom_layer = lambda g: bool_layer_array(g,n=0)
top_layer = lambda g: bool_layer_array(g,n=1)

def get_valley_layer(self,n=0,**kwargs):
    """Get the valley operator for a specific layer"""
    ht = self.copy() # create a dummy
    fac = bool_layer_array(self.geometry,n=n) # create array
    ht.geometry.sublattice = self.geometry.sublattice * fac
    return get_valley(ht,**kwargs) # return the valley operator

operator_list = ["None","Sx","Sy","Sz","valley","sublattice","Berry","valleyberry","IPR","electron","hole","Bulk","Surface","xposition","yposition","zposition"]

def get_layer(self,n=0):
   fac = bool_layer_array(self.geometry,n=n)
   inds = range(len(fac)) # sites
   d = len(fac)
   m = csc_matrix((fac,(inds,inds)),shape=(d,d),dtype=np.complex_) # matrix
   return self.spinless2full(m)




def get_up(self):
    """Return up sector"""
    op = get_sz(self)
    return (op@op + op)/2.


def get_dn(self):
    """Return up sector"""
    op = get_sz(self)
    return (op@op - op)/2.


def get_potential(self,**kwargs):
    """Return the operator associated to a potential"""
    h = self.copy()
    h.clean() # clean the Hamiltonian
    from . import potentials
    f = potentials.commensurate_potential(h.geometry,amplitude=1.0,**kwargs)
    h.add_onsite(f)
    return Operator(h.intra) # return the operator



def get_location(self,r=[0.,0.,0.]):
    "Return the closest index to a specific location"
    ir = self.geometry.closest_index(r)
    return index(self,n=[ir])


def get_site(H,index=0):
    """Return a projector operator in site number index"""
    from . import potentials
    Hp = H*0. # make a dummy copy of the Hamiltonian
    f = potentials.impurity(H.geometry.r[index],v=1.0) # create a local pot
    Hp.add_onsite(f) # add onsite energy
    return Operator(Hp.intra) # return an operator object


def get_correlator_ij(H,i=0,j=0):
    """Return a projector for the correlator ij"""
    d = len(H.geometry.r)
    m = np.zeros((d,d),dtype=np.complex_)
    m[i,j] = 1.0 # set this to nonzero
    # WARNING, this uses dense matrices
    m = H.spinless2full(m) # as full matrix
    return Operator(m) # return operator






import numpy as np


# routines to create selfenergies for different types of leads

class Selfenergy():
    """Self-energy object"""
    def __init__(self,m=None): # pass
        self.m = m
    def __call__(self,**kwargs):
        return self.m



def metal(H,potential):
    """Return the selfenergy of a metallic lead"""
    # create a selfenergy in a single edge (the left one) to kill magnetism
    hs = H.copy()*0.
    hs.add_onsite(lambda r: np.abs(potential(r))) # absolute value
    selfe = -1j*hs.intra # selfenergy on edge, beware of minus sign
    return selfe


def semimetal(*args,zero=0.):
    selfe = metal(*args) # selfenergy of a metal
    def f(energy=0.,**kwargs):
        return selfe*np.abs(energy-zero) # selfenergy of a semimetal
    return f



def get_selfenergy_from_potential(*args,mode="metal",**kwargs):
    if mode=="metal": return metal(*args,**kwargs)
    elif mode=="semimetal": return semimetal(*args,**kwargs)
    else: raise





# routines to call a function in parallel
from __future__ import print_function
import scipy.linalg as lg
from . import algebra


is_child = False # check if you are running a child

try:
    from multiprocess import Pool
    import multiprocess
    maxcpu = multiprocess.cpu_count()
except:
#    print("Multiprocess not working")
    def Pool(n=1): # workaround
            class mpool():
                def map(self,f,xs):
                  return [f(x) for x in xs]
                def terminate(self): return None # dummy function
            return mpool()
    maxcpu = 1


cores = 1 # call in a single by default


def set_cores(n=1):
    global cores
    if n=="max": cores = maxcpu # set to the maximum
    else: cores = int(n) # number of cores


#mainpool = None

#def initialize(): 
#  global mainpool
#  if cores>1:
#    mainpool = Pool(cores) # create pool
#  return mainpool

#def finish(): mainpool=None # delete pool


def multieigh(ms):
  """Diagonalize a bunch of Hamiltonians at once"""
#  mainpool = initialize()
  if mainpool is not None: mainpool.map(lg.eigh,ms)
  else: return [algebra.eigh(m) for m in ms]





def pcall_serial(fun,args):
  """Function to call in serial"""
  return [fun(a) for a in args]


#def pcall_mp(fun,args,cores=1): return pcall_serial(fun,args)

#try: # try to use the multiprocessing library
#  from pathos.multiprocessing import Pool
def pcall_mp(fun,args,cores=cores):
    """Calls a function for every input in args"""
    mainpool = Pool(cores) # create pool
#    print("Using",cores,"cores")
    out = mainpool.map(fun,args) # return list
    mainpool.terminate() # clear the pool
    del mainpool # delete pool
    return out
#except:
#  print("Multiprocessing not found, running in a single core")
#  def pcall_mp(fun,args,cores=1): return pcall_serial(fun,args)
  





def pcall(fun,args): # define the function
  global cores,is_child
  from multiprocessing import current_process
  if not is_child: # if main process
#  if current_process().name == 'MainProcess': # main process
    is_child = True # child from now on
    if cores==1: out = pcall_serial(fun,args) # one core, simply iterate
    else: 
        try: out = pcall_mp(fun,args,cores=cores) # call in parallel
        except:
            print("Something wrong happened in the parallel execution")
            out = pcall_serial(fun,args) # serial execution
    is_child = False # main from now on
    return out
  # child process
  else: return pcall_serial(fun,args) # one core, simply iterate




import numpy as np
from ..htk import fusion


def H2HFH(h,JK=0.0,J=0.):
    """Given a certain geometry, generate a new geometry with 
    heavy fermion sites"""
    if h.has_eh:
        print("Not implemented with superconductivity")
        raise
    if h.has_kondo:
        print("This Hamiltonian already has Kondo sites")
        raise
    width = np.max(h.geometry.r[:,2]) - np.min(h.geometry.r[:,2])
    if width>1e-4: 
        print("Not implemented for non-2D Hamiltonians")
        raise # not implemented for not 2D 
    h.turn_spinful() ; h.turn_dense()
    g = h.geometry # get the geometry
    gl = g.copy() ; gl.r[:,2] += 0.5 ; gl.r2xyz()
    hl = gl.get_hamiltonian(has_spin=True,tij = [J])
    hl.set_filling(0.5) # set to half filling
    # Hamiltonian of the extended modes
    gd = g.copy() ; gd.r[:,2] -= 0.5 ; gd.r2xyz()
    h.geometry = gd
    hd = h.copy() 
    # now the interlayer coupling
    gt = gd + gl
    h = fusion.hamiltonian_fusion(hd,hl)
    h.has_kondo = True # now with Kondo sites
    import types
    h.add_kondo = types.MethodType(add_kondo,h) # add the new method
    if JK!=0.: add_kondo(h,JK) # add Kondo coupling
    return h


def add_kondo(self,JK):
    """Add the Kondo coupling"""
    from ..specialhopping import obj2callable
    JKf = obj2callable(JK) # so that it is callable
    def ft(r1,r2):
        if r1[2]*r2[2]<0.: # different layers
            dr = r1-r2
            rm = (r1+r2)/2. # average position
            dr2 = dr.dot(dr)
            if 0.9<dr2<1.1: return JKf(rm)
        return 0.
    gt = self.geometry
    ht = gt.get_hamiltonian(has_spin=True,tij=ft)
    self.add_hamiltonian(ht) # include this contribution





def get_operator(self,name,**kwargs):
    """Return operators for systems with Kondo sites"""
    if name=="kondo_sites":
        return self.get_operator(lambda r: 1.*(r[2]>0.))
    elif name=="dispersive_electrons":
        return self.get_operator(lambda r: 1.*(r[2]<0.))
    else: raise



def add_onsite(self,ons):
    if not self.has_kondo: raise # only for Kondo Hamiltonians
    m = self.copy() # copy Hamiltonian
    dis = m.get_operator("dispersive_electrons") # operator for dispersive
    intra = m.add_onsite(ons).intra@dis # add onsite only to disp electrons
    self.intra = self.intra + intra # include contribution





from __future__ import print_function
import numpy as np
from scipy import interpolate




def diagram2d(getquantity,x=np.linspace(0.,1.,20),
               y=np.linspace(0.,1.,20),filename="PHASE_DIAGRAM.OUT",
               nite=4):
  """Evaluates a phase diagram, writting the quantities in a file"""
  mz = np.zeros((len(x),len(y))) # 2d array
  mx = np.array([[ix for iy in y] for ix in x]) # 2d array
  my = np.array([[iy for iy in y] for ix in x]) # 2d array
  for i in range(len(x)): # loop over x
    for j in range(len(y)): # loop over x
      q = getquantity(mx[i,j],my[i,j]) # get the value
      mz[i,j] = q # store value

  # now interpolate and call
  mx,my,mz = selected_interpolation(getquantity,mx,my,mz,nite=nite) # interpolate

  fo = open(filename,"w") # open file
  for i in range(len(mx)): # loop over x
    for j in range(len(my)): # loop over x
      fo.write(str(mx[i,j])+"    "+str(my[i,j])+"    "+str(mz[i,j])+"\n") # write in file
  fo.close() # close file




def boundary2d(getquantity,xlim=[0.,1.],
               ylim=[0.,1.],filename="BOUNDARY_PHASE_DIAGRAM.OUT",
               nite=4,tol=0.1,np=50):
  x = np.linspace(xlim[0],xlim[1],np)
  y = np.linspace(ylim[0],ylim[1],4)
  for ix in x: # loop over s parameter
    for iy in y: # loop over s parameter
      raise



def selected_interpolation(fin,mx,my,mz,nite=3):
  """Call in a function in a mesh which is two nite times finer, but only
  in those points that are expected to change, in the others interpolate"""
  if nite<1.01: return mx,my,mz 
  f = interpolate.interp2d(mx[:,0], my[0,:], mz, kind='linear') # interpolation
  x = np.linspace(np.min(mx),np.max(mx),len(mx)*2) # twice as big
  y = np.linspace(np.min(my),np.max(my),len(my)*2) # twice as big
  mx2, my2 = np.meshgrid(x, y) # new grid
  mz2 = f(x,y) # interpolate
  mz2 = mz2.transpose()
  dmz = np.gradient(mz2) # gradient
  dmz = dmz[0]*dmz[0] + dmz[1]*dmz[1] # norm of the derivative
  maxdm = np.max(dmz) # maximum derivative
  for i in range(len(mx2)):
    for j in range(len(my2)):
      if dmz[i,j]>0.001*maxdm: # if derivative is large in this point
        mz2[i,j] = fin(mx2[i,j],my2[i,j]) # re-evaluate the function
  mx2 = mx2.transpose()
  my2 = my2.transpose()
  mz2 = mz2.transpose()
  if nite>1:
    return selected_interpolation(fin,mx2,my2,mz2,nite=nite-1)
  return mx2,my2,mz2 # return function








# from . import geometry
# from . import hamiltonians
# from . import ldos
# from . import embedding
# from . import gap
# from . import massive_green
# from . import operators
# from . import wannier
# from . import topology
# from . import spectrum
# from . import ribbonizate
# from . import multiterminal
# from . import interactions
# from . import green
# from . import hall
# from . import kpm
# from . import kdos
# from . import current
# from . import dos
# from . import sculpt
# from . import kpm
# from . import supercell
# from . import skeleton
# from . import multilayers
# from . import hybrid
# from . import junctions
# from . import multicell



from . import specialhopping
from . import specialgeometry
import numpy as np
from . import geometry


def twisted_multilayer_graphene(n=7,ti=0.12,lambi=8.0,lamb=12.0,is_sparse=True,
        g=None,g0=None,has_spin=False,dl=3.0,**kwargs):
    """
    Return the Hamiltonian of twisted bilayer graphene
    """
    if g is None: g = specialgeometry.twisted_bilayer(n,dz=dl,g=g0,**kwargs)
    mgenerator = specialhopping.twisted_matrix(ti=ti,
            lambi=lambi,lamb=lamb,dl=dl)
    h = g.get_hamiltonian(is_sparse=is_sparse,has_spin=has_spin,
            is_multicell=True,mgenerator=mgenerator)
    return h

# some aliases
twisted_bilayer = twisted_multilayer_graphene
tbg = twisted_bilayer
twisted_bilayer_graphene = tbg

def tdbg(n=4,**kwargs): 
    """Twisted double bilayer Hamiltonian"""
    g = specialgeometry.tdbg(n) # TBDG geometry
    return twisted_multilayer_graphene(g=g,**kwargs) # return Hamiltonian


def multilayer_graphene(l=[0],real=False,**kwargs):
  """Return the hamiltonian of multilayer graphene"""
  g = specialgeometry.multilayer_graphene(l=l)
  g.center()
  if real: # real tight binding hopping
      mgenerator = specialhopping.twisted_matrix(**kwargs)
      h = g.get_hamiltonian(has_spin=False,
          mgenerator=mgenerator)
  else:
      h = g.get_hamiltonian(has_spin=False,
          tij=specialhopping.multilayer(**kwargs))
  return h


def flux2d(g,n=1,m=1):
    """Return a Hamiltonian with a certain commensurate flux per unit cell"""
    from . import sculpt
    from . import supercell
    g = supercell.target_angle(g,0.5) # target a orthogonal cell
    g = sculpt.rotate_a2b(g,g.a1,np.array([1.,0.,0.])) # set in the x direction
    g = g.supercell([1,n,1])
    h = g.get_hamiltonian(has_spin=False)
    r = np.array([0.,1.,0.])
    dr = g.a2.dot(r) # distance
    h.add_peierls(m*2.*np.pi/dr)
    return h



def valence_TMDC(g=None,soc=0.0,**kwargs):
    """Return the Hamiltonian for the valence band of a TMDC"""
    if g is None:
        g = geometry.triangular_lattice()
    ft = specialhopping.phase_C3(g,phi=soc,**kwargs)
    h = g.get_hamiltonian(tij=ft,is_multicell=True,has_spin=False)
    h.turn_spinful(enforce_tr=True)
    return h # return the Hamiltonian


def SOC_TMDC(g=None,soc=0.0,**kwargs):
    """Return the Ising SOC for a triangular lattice"""
    if g is None:
        g = geometry.triangular_lattice()
    ft0 = specialhopping.phase_C3(g,phi=.5,**kwargs)
    ft = lambda r1,r2: 1j*(ft0(r1,r2).imag) # only imaginary part
    h = g.get_hamiltonian(tij=ft,is_multicell=True,has_spin=False)
    h.turn_spinful(enforce_tr=True)
    return h # return the Hamiltonian


from .specialhamiltoniantk.tmdc import doped_MoS2
from .specialhamiltoniantk.heavyfermion import H2HFH




def NbSe2(**kwargs):
    return TMDC_MX2(**kwargs,
            tij=[0.02126916,  0.08462639,  0.00504899, -0.00906057, -0.00572983])

def TaS2(**kwargs):
    return TMDC_MX2(**kwargs,
             tij=[ 0.05525963, 0.08784492, -0.00185766, -0.01204535 ,0.00220995])


def TaS2_SOC(**kwargs):
    return TMDC_MX2(**kwargs,soc=-0.5451483039554,
             tij=[0.05901736,0.10243251,-0.00172311,-0.00993678,0.00459015])



def TMDC_MX2(soc=0.0,cdw=0.0,g=None,tij=[1.0],
              drcdw = np.array([0.,0.,0.]), # shift in the CDW profile
              normalize=True):
    """Return the Hamiltonian of NbSe2"""
    if g is None: 
        g = geometry.triangular_lattice()  # triangular lattice
        if cdw!=0.0: g = g.get_supercell(3,store_primal=True)
    tij = np.array(tij)
    if normalize: tij = tij/np.max(np.abs(tij)) # normalize
    h = g.get_hamiltonian(is_multicell=True,has_spin=False,tij=tij)
    ## Now add the SOC if necessary
    if soc!=0.0: # add the SOC
        h.turn_spinful() # turn spinful
        d = g.neighbor_distances()[1]
        d = 1.0
        hsoc = tij[0]*soc*SOC_TMDC(g=h.geometry,d=d) # hamiltonian with SOC
        h = h + hsoc # add the two Hamiltonians
    if cdw!=0.0: # add the CDW
        g0 = geometry.triangular_lattice()  # triangular lattice
        g0 = g0.supercell(3) # create a supercell
        from . import potentials
        f0 = potentials.commensurate_potential(g0,n=6,angle=np.pi/6.,
                k=1./np.sqrt(3)*2.)
        f0 = f0.normalize()*cdw
        f0 = f0.set_average(0.)
        rc = np.mean(h.geometry.get_closest_position([.1,.1,0.],n=3),axis=0)
        rc = rc - np.array(drcdw) # add the shift
        f = lambda r: f0(r-rc)
        h.geometry.write_profile(f)
        h.add_onsite(f)
    h.set_filling(.5)
#    h = h.supercell(4)
#    m = np.array(h.intra.todense()).reshape(h.intra.shape[0]**2)
    return h
    



def triangular_pi_flux(g=None,**kwargs):
    """Return a pi-flux Hamiltonian"""
    if g is None: # no geometry given
        g = geometry.triangular_lattice() # geometry of a triangular lattice
    g = g.supercell((2,1)) # create a supercell with 2 atoms
    ft = specialhopping.phase_C3_matrix(g,phi=.5)
    h = g.get_hamiltonian(mgenerator=ft,is_multicell=True,**kwargs) # return the Hamiltonian
    h.add_peierls(2*np.pi*1./np.sqrt(3.),gauge="Landau") # field
    from . import gauge
#    h = gauge.hamiltonian_gauge_transformation(h,[0.,0.25])
#    return h
    if h.has_time_reversal_symmetry(): pass
    else: 
        print(np.round(h.intra,2))
        for t in h.hopping:
          print(np.round(t.m,2))
        print("Something wrong happened in pi-flux")
        raise
    exit()
    return h



def excitonic_bilayer(gap=0.0,g=None,**kwargs):
    """Return the Hamiltonian for a bilayer system for an excitonic system"""
    from .geometry import get_geometry
    g0 = get_geometry(g) # get the geometry
    h0 = g0.get_hamiltonian(has_spin=False) # spinless
    hv = h0.copy() # valence
    hc = h0.copy() # conduction
    fermiv = hv.get_fermi4filling(1.0,nk=40) # fermi for valence
    fermic = hc.get_fermi4filling(0.0,nk=40) # fermi for conduction
    gv = g0.copy() # valence geometry
    gc = g0.copy() # conduction geometry
    gv = gv + np.array([0.,0.,-1]) # valence
    gc = gc + np.array([0.,0.,1]) # conduction
    g = gv + gc # add geometries
    def tij(r1,r2):
        dr = r1 - r2
        dr2 = dr.dot(dr)
        if 0.99<dr2<1.01: return np.sign(r1[2])
        return 0.0
    h = g.get_hamiltonian(tij=tij,has_spin=False) # no spin
    def fons(r):
        if r[2]>0.0: return -fermiv - gap/2.
        else: return fermiv + gap/2
    h.add_onsite(fons)
    return h



def FeSe_GXY(nem=0.,**kwargs):
    """Return the Hamiltonian of FeSe, a bandstructure
    displaying two pockets"""
    g = geometry.square_lattice() # cubic lattice
#    h = g.get_hamiltonian(tij=[1.,3.],**kwargs) 
    h = g.get_hamiltonian(tij=[1.,-0.4,0.4],**kwargs) 
#    h.add_onsite(-5.)
    h.add_onsite(-2.5)
    v = np.array([1.,0.,0.]) # nematic vector
    v = v/np.sqrt(v.dot(v))
#    h.remove_spin()
    def fnem(dr):
        if dr.dot(dr)>1e-2: dr = dr/np.sqrt(dr.dot(dr)) # unitary vector
        else: return 1.0 # same site
        o = 1. - nem*(dr.dot(v)**2-0.5) # return 
        return o
    if np.abs(nem)>0.:
      h.add_strain(fnem, mode="directional")
    h.turn_dense()
    return h


FeSe = FeSe_GXY


def FeSe_GM(nem=0.,**kwargs):
    """Return the Hamiltonian of FeSe, a bandstructure
    displaying two pockets, one at G and one at M"""
    g = geometry.square_lattice() # cubic lattice
    h = g.get_hamiltonian(tij=[1.,3.],**kwargs) 
    h.add_onsite(-2.5)
    if nem==0.: return h
    v = np.array([1.,0.,0.]) # nematic vector
    v = v/np.sqrt(v.dot(v))
#    h.remove_spin()
    def fnem(dr):
        if dr.dot(dr)>1e-2: dr = dr/np.sqrt(dr.dot(dr)) # unitary vector
        else: return 1.0 # same site
        o = 1. - nem*(dr.dot(v)**2-0.5) # return 
        return o
    if np.abs(nem)>0.:
      h.add_strain(fnem, mode="directional")
    h.turn_dense()
    return h


from .specialhamiltoniantk.highorderTI import square_2OTI
from .specialhamiltoniantk.highorderTI import chain_1OTI





import numpy as np

def target_moduli(v1,v2,m):
    """Given two vectors, obtain a linear combination with a certain moduli"""
    n = 2*int(m) # estimate of the number
    n = 1
    for i in range(-n,n+1):
        for j in range(-n,n+1):
            v = i*v1 + j*v2
            if abs(np.sqrt(v.dot(v))-m)<1e-3: return v
    raise # not found

def closest_kreplica(g,v,v0):
    """Return the closest replica of a kpoint close to another one"""
    if g.dimensionality==0: return v*0. # for 0d
    elif g.dimensionality==1: # for 1d
        w = v[0]*g.b1 # redefine
        w0 = v0[0]*g.b1 # redefine
        n = 2
        dmax = 1e8
        for i in range(-n,n+1):
                wt = w + g.b1*i # compute replica
                dw = wt - w0 # distance to reference point
                dw = np.sqrt(dw.dot(dw)) # distance
                if 1e-4<dw<dmax: 
                    vo = v + np.array([i,0.,0.])
                    dmax = dw
        return vo # return vo
    elif g.dimensionality==2: # for 2d
        w = v[0]*g.b1 + v[1]*g.b2 # redefine
        w0 = v0[0]*g.b1 + v0[1]*g.b2 # redefine
        n = 2
        dmax = 1e8
        for i in range(-n,n+1):
            for j in range(-n,n+1):
                wt = w + g.b1*i + g.b2*j # compute replica
                dw = wt - w0 # distance to reference point
                dw = np.sqrt(dw.dot(dw)) # distance
                if 1e-4<dw<dmax: 
                    vo = v + np.array([i,j,0.])
                    dmax = dw
        return vo # return vo
    else: raise




def same_kpoint(k1,k2,tol=1e-6):
    """Check if two kpoints are the same"""
    dk = np.array(k1) - np.array(k2) # difference
    dk = np.abs(dk) # absolute value
    dk = np.exp(1j*dk*np.pi*2.) # to the unit circle
    if np.max(np.abs(dk-1.))>tol: return False
    else: return True






def closest_path(g,kps):
    """Given a list of kpoints, return the closest path"""
    out = [kps[0]] # first point
    for i in range(len(kps)-1): # loop over kpoints
        k = closest_kreplica(g,kps[i+1],out[i])
        out.append(k) # store
    return out



def k2path(g,kp,nk=100):
    """Given a set of kpoints, return the path"""
    kp = closest_path(g,kp) # reduced path
    kp = [np.array(k) for k in kp] # convert to arrays
    ks = [] # empty list of kpoints
    for i in range(len(kp)-1): # loop over pairs
      dk0 = kp[i+1] - kp[i] # difference
      dk = dk0[0]*g.b1 + dk0[1]*g.b2 # real reciprocal vector
      dk2 = dk.dot(dk) # moduli real vector
      nk2 = int(nk*np.sqrt(dk2/g.b1.dot(g.b1))) # in units of reciprocal vector
      steps = np.linspace(0.,1.,nk2,endpoint=False) # number of points
      for s in steps:
        ks += [kp[i] + dk0*s] # add kpoint
    ks += [kp[len(kp)-1]] # add the last one
    return ks




from .labels import label2k
from .locate import same_kpoint
import numpy as np

def write_bandlines(g,ks,labels):
    """Given a set of kpoints and labels, write the bandlines in a file"""
    f = open("BANDLINES.OUT","w")
    KHS = [label2k(g,l) for l in labels] # high symmetry points
    def check(k):
        for j in range(len(labels)):
            if same_kpoint(k,KHS[j]): return labels[j] # same kpoint
        return None
    f = open("BANDLINES.OUT","w")
    for i in range(len(ks)): # loop over kpoints
        c = check(ks[i]) # check
        if c is not None: # check
            f.write(str(i)+"  "+str(c)+"\n")
    f.close() # close file







def write_bandlines_indexes(iks,labels):
    """Write a file with the special kpoints"""
    f = open("BANDLINES.OUT","w")
    for (ik,l) in zip(iks,labels):
        f.write(str(ik)+"  "+str(l)+"\n")
    f.close()






# routines to update the mean field
from .. import meanfield
from .. import check
import numpy as np
from numba import jit

def sum_contributions(a,b,vav,vbv,g,inds):
    """Compute all the contributions to the mean field"""
    dirs = [] # empty list


def generate_index2vec(self):
    """Generate the index2vec variables"""
    if self.has_index2vec: return # already generated
    ds = [v.dir for v in self.interactions] # directions
    uniqued = []
    for d in ds:
        if not tuple(d) in uniqued: uniqued.append(tuple(d))
    inds = []
    for v in self.interactions: # loop
        i = uniqued.index(tuple(v.dir)) # get the index
        inds.append(i) # store
    self.interactions_dindex = np.array(inds) # store
    self.interactions_vecs = uniqued # store
    self.has_index2vec = True # generated


def update_mean_field(self,mixing=0.95):
    """Calculate the expectation values of the different operators"""
    generate_index2vec(self) # generate the array of not present yet
    accu = 0.0 # accumulator
    if self.correlator_mode == "multicorrelator":
      np.savetxt("VS_SCF.OUT",np.matrix([range(len(self.cij)),np.abs(self.cij)]).T)
    zero = self.hamiltonian.intra*0.
    # initialize
    storage = [zero.copy() for i in range(len(self.interactions_vecs))]
    for ii in range(len(self.interactions)): # loop over interactions
      v = self.interactions[ii] # get this one
      if v.contribution=="AB":
        tmp = v.a*v.vbv*v.g + v.b*v.vav*v.g # add contribution
        accu += np.abs(v.vbv) + np.abs(v.vav)
      elif v.contribution=="A":
        tmp = v.a*v.vbv*v.g
        accu += np.abs(v.vbv)
      else: raise
      # store in the dictionary
      jj = self.interactions_dindex[ii] # index for this one
      storage[jj] = storage[jj] + tmp # store
    ##################################
    ##################################
    # now put it in a dictionary
    mfnew = dict() # new mean field
    jj = 0
    for d in self.interactions_vecs: # get this one
      mfnew[tuple(d)] = storage[jj] # add contribution
      jj += 1
    accu /= len(self.interactions)*2
    if not self.silent: print("Average value of expectation values",accu)
    self.error = error_meanfield(self.mf,mfnew)/self.sites # get the error
    self.mf = mix_meanfield(self.mf,mfnew,mixing=mixing) # new mean field
    if self.hamiltonian.has_eh:
      print("################################")
      print("Enforcing electron-hole symmetry")
      print("################################")
      self.mf = meanfield.enforce_eh(self.hamiltonian,self.mf)
    if not self.silent: print("ERROR",self.error)
    check.check_dict(self.mf)




def error_meanfield(mf1,mf2):
  """Return the difference between two mean fields"""
  error = 0.0 # initialize
  for key in mf1:
    error += np.sum(np.abs(mf1[key]-mf2[key])) # sum error
  return error



def mix_meanfield(mf1,mf2,mixing=0.9):
  """Mix the two mean fields"""
  out = dict() # create dictionary
  for key in mf2:
    if key in mf1: # if present in the old one
      out[key] = mf1[key]*(1-mixing)+mf2[key]*mixing # sum error
    else:
      out[key] = mf2[key] # sum error
  return out





# library for creating different kinds of junctions

import numpy as np
from . import geometry  # library with different geometries 
from . import hamiltonians  # hamiltonians library

class junction():
  def __init__(self):
    self.number = 3  



def generator_super_squid_square(width=1,inner_radius=4,arm_length=8,
          arm_width=2,delta=0.0,mag_field=0.0,fill=False):
  """ Creates a function that returns a superconducting 
     squid made of square lattice,
     central part is a normal system"""
  g = geometry.squid_square(width=width,inner_radius=inner_radius,
                              arm_length=arm_length, 
                              arm_width = arm_width,fill=fill)  # create the geometry
  g.dimensionality = 0 # set 0 dimensional
  g.celldis = None # no distance to neighboring cell
  h = hamiltonians.hamiltonian(g) # create hamiltonian
  nt = width + inner_radius+1. # half side of the big square
  def mag_field_f(x1,y1,x2,y2):
    """ Return magnetic field only in the center,
       using atan x gauga """
  #  return mag_field*(x1-x2)*(y1+y2)/2.0
  #  if (x1 or x2) < -float(nt): return 0.0
  #  if (x1 or x2) > float(nt): return 0.0
  #  else: return mag_field*(x1-x2)*(y1+y2)/2.0
    bb = (np.arctan(x1/nt)+np.arctan(x2/nt))/2.0  # mean x value
    bb = bb*(y2-y1)*mag_field
    return bb
  h.get_simple_tb(mag_field=mag_field_f)  # create simple tight binding
  def delta_f(x,y):
    """ Defines left and right arms arm """
    if x < -float(nt): return delta
    if x > float(nt): return delta
    else: return 0.0
  def h_generator(phi):
    """ Creates the function to return"""
    def phi_f(x,y):
      """ Defines left and right arms arm """
      if x < -float(nt): return phi
      if x > float(nt): return -phi
      else: return 0.0
    # add superconducting parameter
    h.remove_spin() # remove spin degree of freedom
    h.add_swave_electron_hole_pairing(delta=delta_f,phi=phi_f)
    h.set_finite_system() # transform to finite system
    return h # return hamiltonian
  return h_generator



def super_normal_super(g,f_sc,f_n,replicas = [1,1,1],phi=0.0,delta=0.1):
  """Creates a junction superconductor-normal-superconductor,
          - g is the geometry of the unit cell of the system
          - f_sc is the function which generates the superconductor
            which takes three arguments, geometry, SC order and phase
          - f_n generates the hamiltonian of the central part,
            result has to have electron-hole dimension
          - replicas are the replicas of the SC-normal_SC of the cell
          - phi is the infinitesimal phase difference
          - delta is the superconducting parameter """
  h_left = f_sc(g,delta=delta,phi=phi) # create hamiltonian of the left part
  h_right = f_sc(g,delta=delta,phi=-phi) # create hamiltonian of the right part
  h_central = f_n(g)  # central part
  from scipy.sparse import coo_matrix as coo
  from scipy.sparse import bmat
  from hamiltonians import hamiltonian
  hj = hamiltonian(g) # create hamiltonian of the junction
  tc = replicas[0] + replicas[1] + replicas[2] # total number of cells
  intra = [[None for i in range(tc)] for j in range(tc)]  # create intracell


  # transfor to coo matrix
  h_left.intra = coo(h_left.intra)
  h_left.inter = coo(h_left.inter)
  h_central.intra = coo(h_central.intra)
  h_central.inter = coo(h_central.inter)
  h_right.intra = coo(h_right.intra)
  h_right.inter = coo(h_right.inter)


  # intracell contributions
  for i in range(replicas[0]): 
    intra[i][i] = h_left.intra # intracell of the left lead
  init = replicas[0] # inital index
  for i in range(replicas[1]): 
    intra[init+i][init+i] = h_central.intra # intracell of the left lead
  init = replicas[0]+replicas[1] # inital index
  for i in range(replicas[2]): 
    intra[init+i][init+i] = h_right.intra # intracell of the left lead

  # intercell contributions
  for i in range(replicas[0]-1): 
    intra[i][i+1] = h_left.inter # intercell of the left lead
    intra[i+1][i] = h_left.inter.H # intercell of the left lead
  init = replicas[0] # inital index
  for i in range(replicas[1]-1): 
    intra[init+i][init+i+1] = h_central.inter # intercell of the central lead
    intra[init+i+1][init+i] = h_central.inter.H # intercell of the central lead
  init = replicas[0]+replicas[1] # inital index
  for i in range(replicas[2]-1): 
    intra[init+i][init+i+1] = h_right.inter # intercell of the right lead
    intra[init+i+1][init+i] = h_right.inter.H # intercell of the right lead

  # coupling between SC and central part, take couapling of the central
  il = replicas[0]
  intra[il-1][il] = h_central.inter # intracell of the left lead
  intra[il][il-1] = h_central.inter.H # intracell of the left lead
  il = replicas[0]+replicas[1]
  intra[il-1][il] = h_central.inter # intracell of the left lead
  intra[il][il-1] = h_central.inter.H # intracell of the left lead
  # create matrix
  intra = bmat(intra).todense()
  hj.intra = intra # add to hamiltonian junction
  hj = hj.finite_system() # convert in finite system
  return hj




import numpy as np
from .. import algebra
import scipy.sparse.linalg as slg
from scipy.sparse import csc_matrix


arpack_tol = algebra.arpack_tol
arpack_maxiter = algebra.arpack_maxiter


def states_generator(h,filt=None,max_waves=None):
    hkgen = h.get_hk_gen() # get hamiltonian generator
    def agen(k):   
        hk = hkgen(k) # get hamiltonian
        if max_waves is None: 
            es,wfs = algebra.eigh(hk) # diagonalize all waves
        else:  
            es,wfs = slg.eigsh(csc_matrix(hk),k=max_waves,which="SA",
                        sigma=0.0,tol=arpack_tol,maxiter=arpack_maxiter)
#        es,wfs = algebra.eigh(hk) # diagonalize all waves
        wfs = np.conjugate(wfs).T
        # now filter positive and negative energies
        if filt is not None: wfs = filt(wfs,k=k) # filter wavefunctions
        return wfs
    return agen


def positive_wf(es,wfs):
    """Return eigenfunctions with positive eigenvalues"""
    es0 = []
    wfs0 = []
    for (e,wf) in zip(es,wfs):
        if e>0.0: # store
            es0.append(e)
            wfs0.append(wf)
    wfs1 = [wf for (e,wf) in sorted(es0,wfs0)] # sorted wavefunctions
    es1 = sorted(es) # sorted energies
    return np.array(wfs1)






def occ_states_generator(h,k,**kwargs):
    """Return a function that generates the occupied wavefunctions"""
    if not h.os_gen:
        hk_gen = h.get_hk_gen() # get hamiltonian generator
        return lambda k: occupied_states(hk_gen,k,**kwargs)
    else: return h.os_gen




def occupied_states(hkgen,k,window=None,max_waves=None):
    """ Returns the WF of the occupied states in a 2d hamiltonian"""
    hk = hkgen(k) # get hamiltonian
    if max_waves is None: es,wfs = algebra.eigh(hk) # diagonalize all waves
    else:  es,wfs = slg.eigsh(csc_matrix(hk),k=max_waves,which="SA",
                        sigma=0.0,tol=arpack_tol,maxiter=arpack_maxiter)
    wfs = np.conjugate(wfs.transpose()) # wavefunctions
    occwf = []
    for (ie,iw) in zip(es,wfs):  # loop over states
      if window is None: # no energy window
        if ie < 0.:  # if below fermi
          occwf.append(iw)  # add to the list
      else: # energy window provided
        if -abs(window)< ie < 0:  # between energy window and fermi
          occwf.append(iw)  # add to the list
    return np.array(occwf)


def occ_states2d(h,k):
  """Input is a Hamiltonian"""
  hk_gen = h.get_hk_gen() # get hamiltonian generator
  return occupied_states(hk_gen,k)



class Filter():
    """Class for filtering states"""
    def __init__(self,f):
        if type(f)==Filter: self.f = f.f # store
        if callable(f): self.f = f # store
    def __call__(self,*args,**kwargs):
        return self.f(*args,**kwargs)
    def __mul__(self,a):
        a = Filter(a) # convert
        f2 = lambda *args,**kwargs: self(a.f(*args,**kwargs),**kwargs)
        return Filter(f2) # return new function



def filter_state(opk,accept=lambda r: True, nmax = None):
   """Flter certain states according to their eigenvalues"""
   def filt(wfs,k=None):
       n = wfs[0].shape[0]
       iden = np.identity(n,dtype=np.complex_)
       op = opk(iden,k=k) # evaluate at a kpoint if needed
       wfs = algebra.disentangle_manifold(wfs,op) # disentangle
       ls = algebra.get_representation(wfs,op) # get their eigenvalue
       ls = [ls[i,i].real for i in range(len(ls))]
       out = []
       for (l,w) in zip(ls,wfs):
           if accept(l): 
               out.append(w)
       return out # return wavefunctions
   return Filter(filt) # return filter



def max_valence_states(h,n=2):
    """Function to filter a maximum number of valence states"""
    opk = h.get_operator("energy") # energy operator
    def filt(wfs,k=None):
       ni = wfs[0].shape[0]
       iden = np.identity(ni,dtype=np.complex_)
       op = opk(iden,k=k) # evaluate at a kpoint if needed
       wfs = algebra.disentangle_manifold(wfs,op) # disentangle
       ls = algebra.get_representation(wfs,op) # get their eigenvalue
       ls = [ls[i,i].real for i in range(len(ls))]
       out = negative_wf(ls,wfs) # stored wavefunctions
       if len(out)>=n: out = out[0:n,:]
       else:
           print("Not enough wavefunctions",len(out),n)
       return out # return wavefunctions
    return Filter(filt)


def negative_wf(es,wfs):
    """Return eigenfunctions with positive eigenvalues"""
    es = np.array(es)
    es0 = []
    wfs0 = []
    for (e,wf) in zip(es,wfs):
        if e<0.0: # store
            es0.append(e)
            wfs0.append(wf)
    es0 = np.array(es0)
    # sorted wavefunctions
    wfs1 = [wf for (e,wf) in sorted(zip(-es0,wfs0),key=lambda x: x[0])] 
    es1 = -np.sort(-es) # sorted energies
    return np.array(wfs1)





import numpy as np
from scipy.integrate import quad
from scipy.interpolate import interp1d
from ..transporttk.fermidirac import fermidirac as FD


def temperature_convolution(x,y,temp=1e-1):
    if temp==0.: return x,y
    dt = 20 # hwo many times the temperature
    x = np.array(x)
    y = np.array(y)
    fy = interp1d(x,y,bounds_error=False,fill_value=(y[0],y[len(y)-1]))
    de = temp/10
    def g(e0):
        def f(e):
            out = fy(e+e0) # call the function
            out *= FD(e-de,temp=temp) - FD(e+de,temp=temp)
            return out/de
        return quad(f,-dt*temp,dt*temp,epsrel=1e-2,limit=30)[0]/2.
    return x,np.array([g(ix) for ix in x]) # return




import numpy as np

def randomwf(nd):
    """Return a random wavefunction of simension n"""
    def fun():
      v = (np.random.random(nd) - 0.5)*np.exp(2*1j*np.pi*np.random.random(nd))
      return v/np.sqrt(np.sum(np.abs(v)**2)) # normalize
    return fun



import os
import numpy as np
from ..scftypes import scfclass
from ..scftypes import get_fermi_energy
from ..scftypes import get_occupied_states
import time
from .. import limits
from .. import inout
from scipy.sparse import csc_matrix



mf_file = "MF.pkl" # mean field file


def coulombscf(h,g=1.0,nkp = 100,filling=0.5,mix=0.9,
                  maxerror=1e-05,silent=False,mf=None,
                  smearing=None,fermi_shift=0.0,
                  maxite=1000,save=False,**kwargs):
  """ Solve a selfconsistent Hubbard mean field"""
  vc = coulomb_density_matrix(h.geometry,vc=2*g,**kwargs) # get the matrix
  mix = 1. - mix
  U = g # redefine
#  if h.has_spin: raise # not implemented
  os.system("rm -f STOP") # remove stop file
  from scipy.linalg import eigh
  nat = h.intra.shape[0] # number of atoms
  htmp = h.copy()  # copy hamiltonian
  htmp.turn_dense() # turn into a dense Hamiltonian
  # generate the necessary list of correlators
  if mf is None: # generate initial mean field
    try:  
        old_mf = inout.load(mf_file) # load the file
        print("Mean field read from file")
    except: # generate mean field
        old_mf = np.random.random((nat,nat))
        old_mf += old_mf.T
  else: old_mf = mf # use guess
  # get the pairs for the correlators
  ndim = nat # dimension
  totkp = nkp**(h.dimensionality) # total number of kpoints
  file_etot = open("SCF_ENERGY.OUT","w")
  file_error = open("SCF_ERROR.OUT","w")
  ite = 0 # iteration counter
  scf = scfclass(h) # create scf class
  while True: # infinite loop
    ite += 1 # increase counter
    htmp.intra = h.intra + old_mf # add mean field 
    t1 = time.time()
# get eigenvectors
    eigvals,eigvecs,kvectors = htmp.get_eigenvectors(nk=nkp,kpoints=True)
    eigvecs = np.conjugate(eigvecs)
# fermi energy
    t2 = time.time()
    fermi = get_fermi_energy(eigvals,filling,fermi_shift=fermi_shift)
# occupied states
    eoccs,voccs,koccs = get_occupied_states(eigvals,eigvecs,kvectors,fermi)
# mean field
    # get the new mean field
    if h.has_spin: mf,edc = charge_mean_field_spinful(voccs,vc) 
    else: mf,edc = charge_mean_field(voccs,vc) # get the new mean field
    mf = mf/totkp # normalize
    edc = edc/totkp # normalize
    t3 = time.time()
    error = np.max(np.abs(old_mf-mf)) # absolute difference
    # total energy
    etot = (np.sum(eoccs)/totkp + edc).real  # eigenvalues and double counting
    file_etot.write(str(ite)+"    "+str(etot)+"\n") # write energy in file
    file_error.write(str(ite)+"    "+str(error)+"\n") # write energy in file
    file_etot.flush()
    file_error.flush()
#    totcharge = np.sum(charge).real # total charge
#    avcharge = totcharge/nat # average charge
    if save: inout.save(mf,mf_file) # save the mean field
    ######
    if not silent:
      print("Times in diagonalization",t2-t1)
      print("Times in new mean field",t3-t2)
      print("\n")
      print("Iteration number =",ite)
      print("Error in SCF =",error)
      print("Fermi energy =",fermi)
      print("Total energy =",etot)
#      print("Total charge =",totcharge)
    old_mf = old_mf*mix + mf*(1.-mix) # mixing
    if error<maxerror or os.path.exists("STOP"): # if converged break
      break
    if ite>=maxite: break # if too many iterations
  file_etot.close() # close file
  file_error.close() # close file
  # output result
  class scftmp(): pass # create an empty class
  scf = scftmp() # empty class
  scf.hamiltonian = htmp.copy() # copy Hamiltonian
  scf.hamiltonian.intra -= fermi*np.identity(ndim) # shift Fermi energy
  scf.total_energy = etot # store total energy
  scf.mf = mf # store mean field matrix
  return scf # return mean field





def charge_mean_field(voccs,vc):
    """Return the charge mean field"""
    n = voccs[0].shape[0]
    charge = np.zeros(n) # initialize
    mf = np.zeros(n) # initialize
    for w in voccs:
        charge += np.abs(w)**2 # add contribution
    for i in range(len(vc)): # loop over rows of the matrix
        mf[i] = charge.dot(vc[i]) # get contribution
    # This formula is only ok if there is a single pair per ij
    edc = -mf@(vc@mf)
    return np.diag(mf),edc # return diagonal matrix



def charge_mean_field_spinful(voccs,vc):
    """Return the charge mean field, for a spinful Hamiltonian
       Vc is the spinless matrix"""
    n = vc.shape[0] # number of sites
    charge = np.zeros(2*n) # initialize
    mfup = np.zeros(n) # initialize
    mfdn = np.zeros(n) # initialize
    mf = np.zeros(2*n) # initialize
    for w in voccs:
        charge += np.abs(w)**2 # add contribution
    chargeup = np.array([charge[2*i] for i in range(n)]) # up charge
    chargedn = np.array([charge[2*i+1] for i in range(n)]) # down charge
    for i in range(n): # loop over rows of the matrix
        mfdn[i] = chargeup.dot(vc[i]) # get contribution
        mfup[i] = chargedn.dot(vc[i]) # get contribution
    # This formula is only ok if there is a single pair per ij
    edc = -mfup@(vc@mfdn)
    for i in range(n):
        mf[2*i] = mfup[i]
        mf[2*i+1] = mfdn[i]
    return np.diag(mf),edc # return diagonal matrix
















def coulomb_density_matrix(g,rcut=30.0,vc=0.0):
    """Return an array with the Hartree terms"""
    g = g.copy() # copy geometry
    interactions = [] # empty list
    nat = len(g.r) # number of atoms
    mout = np.zeros((nat,nat)) # initialize matrix
    lat = np.sqrt(g.a1.dot(g.a1)) # size of the unit cell
    g.ncells = int(2*rcut/lat)+1 # number of unit cells to consider
    ri = g.r # positions
    for d in g.neighbor_directions(): # loop
        rj = np.array(g.replicas(d)) # positions
        for i in range(nat):
          for j in range(nat):
            dx = rj[j,0] - ri[i,0]
            dy = rj[j,1] - ri[i,1]
            dz = rj[j,2] - ri[i,2]
            dr = dx*dx + dy*dy + dz*dz
            dr = np.sqrt(dr) # square root
            if dr<1e-4: v = 0.0
            elif dr>rcut: v = 0.0
            else: v = 1./dr*np.exp(-dr/rcut) # quench interaction
            mout[i,j] += v # store contribution
    mout /= 2*np.max(mout) # normalize
    return vc*mout # return







import numpy as np
from numba import jit

def get_nnc(g,den,n=20,normalized=False,**kwargs): 
    """Compute the first N correlators"""
    g0 = g.copy() ; g0.dimensionality = 0 # zero dimensional
    ds = g0.neighbor_distances() # neighbor distances
    if len(ds)<n: n = len(ds)
    r = g0.r # locations
    ri = r.copy() # copy locations
    deni = den.copy()
    if g.dimensionality>0: # periodic boundaries
        rj = g.multireplicas(1) # replicas of the locations
        den0 = den.copy() # copy
        denj = den0.copy() # start
        for i in range(1,len(rj)//len(den0)): denj = np.concatenate([denj,den0])
    else:
        rj = ri
        denj = deni
    cs = [get_nnci_jit(ri,rj,deni,denj,di,1e-3) for di in ds[0:n]]
    cs = np.array(cs) # convert to array
    if normalized: # normalize using Cauchy-Swartz inequality
        var = np.mean((den - np.mean(den))**2)
        # ignore replica effect due to periodic BC
        cs = cs/var # normalize by variance
    x,y = ds[0:len(cs)],cs # distance and correlators
    return x,y











@jit(nopython=True)
def get_nnci_jit_v1(ri,rj,deni,denj,di,delta):
    """Compute a single correlator"""
    out = 0. # output value
    no = 0. # counter
    outi,outj,noi,noj = 0.,0.,0.,0.
    for i1 in range(len(ri)):
        for i2 in range(len(rj)):
            r1 = ri[i1]
            r2 = rj[i2]
            dr = r1-r2
            dr2 = np.sum(dr*dr)
            if np.abs(dr2 - di**2)<delta:
                out += deni[i1]*denj[i2]
                no += 1.
                outi += deni[i1]
                outj += denj[i2]
                noi += 1
                noj += 1
    if no==0: return 0.
    return out/no  - outi/noi*outj/noj # return correlator








@jit(nopython=True)
def get_nnci_jit(ri,rj,deni,denj,di,delta):
    """Compute a single correlator"""
    out = 0. # output value
    no = 0. # counter
    deni = deni - np.mean(deni) # redefine so that it has zero average
    denj = denj - np.mean(denj) # redefine so that it has zero average
    for i1 in range(len(ri)):
        for i2 in range(len(rj)):
            r1 = ri[i1]
            r2 = rj[i2]
            dr = r1-r2
            dr2 = np.sum(dr*dr)
            if np.abs(dr2 - di**2)<delta:
                out += deni[i1]*denj[i2]
                no += 1.
    if no==0: return 0. # no neighbor found
    return out/no  #- np.mean(deni)*np.mean(denj) # return correlator



# routines to compute magnetic exchange
import numpy as np




def NN_exchange(h,J=0.1,nk=2,num_bands=None,full_energy=False,
        mode="supercell",filling=None):
    """Compute the nearest neighbor exchange using
    a brute force algorithm"""
    if not full_energy:  nbands = num_bands # just the same method for energy
    else: nbands = None # enforce full calculation
    from . import ldos
    d = ldos.ldos_density(h,nk=nk,delta=1e-3,num_bands=num_bands) # density
    if mode=="supercell":
        hs = h.supercell([2,1,1]) # get a supercell
        ds0 = J*np.concatenate([d,d]) # FE configuration
        ds1 = J*np.concatenate([d,-d]) # AF configuration
        if h.has_spin: # spinful version
            ds0 = np.array([0*ds0,0*ds0,ds0]).T # magnetizations
            ds1 = np.array([0*ds1,0*ds1,ds1]).T # magnetizations
            hs0 = hs.copy()
            hs1 = hs.copy()
            hs0.add_zeeman(ds0) # add the Zeeman field
            hs1.add_zeeman(ds1) # add the Zeeman field
            e0 = hs0.get_total_energy(nk=nk,nbands=nbands) # GS energy 
            e1 = hs1.get_total_energy(nk=nk,nbands=nbands) # GS energy 
            return (e1-e0)/2.
        else: # spinless version
            hs0u = hs.copy()
            hs0d = hs.copy()
            hs1u = hs.copy()
            hs1d = hs.copy()
            hs0u.add_onsite(ds0)
            hs0d.add_onsite(-ds0)
            hs1u.add_onsite(ds1)
            hs1d.add_onsite(-ds1)
            gete = lambda h0: h0.get_total_energy(nk=nk,nbands=nbands) 
            return (gete(hs1u) + gete(hs1d) - gete(hs0u) - gete(hs0d))/2.
    elif mode=="spiral": # spin spiral mode
        dm = J*np.array([d,0*d,0*d]).T # magnetizations
        h = h.copy() # copy Hamiltonian
        h.turn_spinful() # add spin
        h.add_zeeman(dm) # add exchange field
        if filling is not None: h.set_filling(filling,nk=nk) # set filling
        h0 = h.copy() # get the ferro
        h1 = h.copy() # get the AF
        h1.generate_spin_spiral(vector=[0,0,1],qspiral=[1,0,0])
        e0 = h0.get_total_energy(nk=nk,nbands=nbands) # GS energy 
        e1 = h1.get_total_energy(nk=nk,nbands=nbands) # GS energy 
        return e1-e0
    else: raise # unrecognized mode






import numpy as np
from numba import jit
import os
from .. import filesystem as fs
from numba import jit


@jit(nopython=True)
def jit_get_orbital(r0,rs,ratomic): 
    """Atomic orbital"""
    dr = rs-r0
    dr2 = np.sum(dr*dr,axis=1) # sum
    return np.exp(-np.sqrt(dr2)/ratomic)


def ldos_generator(h,**kwargs):
    return profile_generator(h,mode="LDOS",**kwargs)


def profile_generator(h,delta=0.05,nrep=1,nk=20,dl=None,mode="LDOS",
        ratomic=1.5,dr=0.2,num_bands=None,
        deltax=None, deltay=None,**kwargs):
    """Compute the LDOS at different eenrgies, and add an envelop atomic
    orbital"""
    h = h.copy() # copy the Hamiltonian
    h.turn_dense() # dense hamiltonian
    evals,vs,ks = h.get_eigenvectors(nk=nk,kpoints=True,
            numw=num_bands) # compute wavefunctions
    if dl is None: 
        if h.dimensionality==0: nrepdl = int(ratomic*10)
        dl = h.geometry.neighbor_directions(nrep+int(ratomic*10)) # directions of the neighbors
    def get_orbital(r0,r):
        return jit_get_orbital(r0,r,ratomic)
    # generate a dictionary with all the real space local orbitals
    ##########################################################
    lodict = dict() # dictionary for the local orbitals
    # get the grids
    if deltax is None: deltax=ratomic
    if deltay is None: deltay=ratomic
    x,y = get_grids(h.geometry,nrep=nrep,dr=dr,
            deltax=deltax,deltay=deltay)
    r = np.zeros((len(x),3)) ; r[:,0] = x ; r[:,1] = y
    # now chack which centers to accept
    xmin,xmax = np.min(x),np.max(x)
    ymin,ymax = np.min(y),np.max(y)
    def accept_center(r):
        """Check if this center is close enough"""
        fac = 10
        if r[0]-xmin<-(fac*ratomic): return False # too left
        if r[0]-xmax>(fac*ratomic): return False # too right
        if r[1]-ymin<-(fac*ratomic): return False # too down
        if r[1]-ymax>(fac*ratomic): return False # too right
        return True
    for d in dl: # loop over directions
          rrep = h.geometry.replicas(d=d) # replicas in this direction
          for i in range(len(rrep)): # loop over the atoms
              r0 = rrep[i] # get this center
              if not accept_center(r0): continue # skip this iteration
              if h.has_eh:
                if h.has_spin: # spinful
                  lodict[(tuple(d),4*i)] = get_orbital(r0,r) # store 
                  lodict[(tuple(d),4*i+1)] = get_orbital(r0,r) # store 
           #       lodict[(tuple(d),4*i+2)] = 0. # store 
           #       lodict[(tuple(d),4*i+3)] = 0. # store 
                else: raise
              else:
                if h.has_spin: # spinful
                  lodict[(tuple(d),2*i)] = get_orbital(r0,r) # store 
                  lodict[(tuple(d),2*i+1)] = get_orbital(r0,r) # store 
                else: # spinless
                  lodict[(tuple(d),i)] = get_orbital(r0,r) # store 
    ##########################################################
    # now compute the real-space wavefunctions including the Bloch phase
    ds = np.zeros((len(vs),len(x))) # zero array
    # get the generator
    density_generator = get_real_space_density_generator(lodict,h.geometry,
                              has_spin=h.has_spin)
    for i in range(len(vs)): # loop over wavefunctions
        w = vs[i] # get the current Bloch wavefunction
        k = ks[i] # get the current bloch wavevector
        d = density_generator(w,k)
        ds[i] = d # store in the list
    if mode=="LDOS": # LDOS mode
      def f(e): return ldos_at_energy(evals,ds,e,delta) # compute the LDOS
    elif mode=="density": # LDOS mode
      def f(e): return density_at_energy(evals,ds,e,delta) # compute the LDOS
    else: raise # not implemented
    return f,evals,x,y # return generator


def get_ldos(h,e=0.0,delta=0.05,**kwargs):
    """Compute a single LDOS"""
    ldos_gen,evals,x,y = ldos_generator(h,e=e,delta=delta,**kwargs) 
    out = ldos_gen(e) # compute the LDOS
    np.savetxt("LDOS.OUT",np.array([x,y,out]).T) # save
    return x,y,out


def get_density(h,e=0.0,delta=1e-3,**kwargs):
    """Compute a single LDOS"""
    ldos_gen,evals,x,y = profile_generator(h,e=e,delta=delta,
                               mode="density",**kwargs)
    out = ldos_gen(e) # compute the LDOS
    np.savetxt("DENSITY.OUT",np.array([x,y,out]).T) # save
    return x,y,out




def multi_ldos(h,es=np.linspace(-2.0,2.0,100),delta=0.05,**kwargs):
    """Compute the LDOS at different eenrgies, and add an envelop atomic
    orbital"""
    ldos_gen,evals,x,y = ldos_generator(h,delta=delta,**kwargs) # get the generator
    # now compute all the LDOS
    fs.rmdir("MULTILDOS")
    fs.mkdir("MULTILDOS")
    fo = open("MULTILDOS/MULTILDOS.TXT","w") # files with the names
    for e in es: # loop over energies
        name0 = "LDOS_"+str(e)+"_.OUT" # name of the output
        name = "MULTILDOS/" + name0
        out = ldos_gen(e) # compute the LDOS
        np.savetxt(name,np.array([x,y,out]).T) # save
        fo.write(name0+"\n") # name of the file
    fo.close()
    from ..dos import calculate_dos,write_dos
    es2 = np.linspace(min(es),max(es),len(es)*10)
    ys = calculate_dos(evals,es2,delta,w=None) # compute DOS
    write_dos(es2,ys,output_file="MULTILDOS/DOS.OUT")



def get_real_space_density_generator(lodict,g,has_spin=False):
    """Compute the orbital in real space"""
    out = 0. # wavefunction in real space
    orbs = np.array([lodict[key] for key in lodict]) # orbitals
    inds = np.array([key[1] for key in lodict],dtype=int) # indexes
    ds = [key[0] for key in lodict]
    def fout(w,k):
        """Function that return the real space density"""
        nc = len(w) # number of components of the Bloch wavefunction
        phis = np.array([g.bloch_phase(d,k) for d in ds]) # phases
        out = np.zeros(orbs[0].shape[0],dtype=np.complex_)
        if not has_spin: # spinless
          return get_real_space_density_jit(w,phis,inds,orbs,out).real
        else: # spinful
          wup = ud_component(w,mode="up") 
          wdn = ud_component(w,mode="dn") 
          outup = get_real_space_density_jit(wup,phis,inds,orbs,out).real
          outdn = get_real_space_density_jit(wdn,phis,inds,orbs,out).real
          return outup + outdn
    return fout # return function

@jit(nopython=True)
def get_real_space_density_jit(w,phis,inds,orbs,out):
    """Return the real space wavefunction"""
    for ii in range(len(inds)):
        iorb = int(inds[ii]) # integer
        out = out + w[iorb]*phis[ii]*orbs[ii] # add contribution
    return out*np.conjugate(out)




def get_grids(g,nrep=1,dr=0.1,deltax=1.0,deltay=1.0):
    """Return the grids to plot the real space wavefunctions"""
    r = g.multireplicas(nrep-1) # get all the position
    xmin = np.min(r[:,0])
    xmax = np.max(r[:,0])
    ymin = np.min(r[:,1])
    ymax = np.max(r[:,1])
    nx = int((xmax-xmin+2*deltax)/dr) # number of x points
    ny = int((ymax-ymin+2*deltay)/dr) # number of y points
    xp = np.linspace(xmin-deltax,xmax+deltax,nx) # generate the points
    yp = np.linspace(ymin-deltay,ymax+deltay,ny) # generate the points
    gridx = np.zeros(nx*ny)
    gridy = np.zeros(nx*ny)
    gridx,gridy = get_grids_jit(xp,yp,gridx,gridy)
    return gridx,gridy # return the grids

@jit(nopython=True)
def get_grids_jit(x,y,gridx,gridy):
    nx = len(x)
    ny = len(y)
    k = 0
    for i in range(nx):
      for j in range(ny):
          gridx[k] = x[i]
          gridy[k] = y[j]
          k += 1
    return gridx,gridy

def ldos_at_energy(evals,ds,e,delta):
    """Compute the different local density of states at each energy"""
    de2 = (evals-e)**2 # difference in energy
    out = np.sum(ds.T*delta/(de2+delta**2),axis=1)
    return out # return that density

def density_at_energy(evals,ds,e,delta):
    """Compute the density at this energy"""
    de = evals-e # difference in energy
    w = (1. - np.tanh(de/delta))/2. # weight
    out = np.sum(ds.T*w,axis=1) # output
    return out # return that density



def ud_component(w,mode="up"):
    """Given a wavefunction, return the up components"""
    n = len(w)//2
    wo = w.copy()+0j # copy
    if mode=="up": p = 0
    elif mode=="dn": p = 1
    else: raise
    for i in range(n): wo[2*i+p] = 0.0j # set to zero
    return wo




import numpy as np
from scipy.sparse import csc_matrix
from numba import jit

use_fortran=False


def obj2callable(a):
    if callable(a): return a # input is a function
    else: return lambda x: a # input is a number



def twisted(cutoff=5.0,ti=0.3,lambi=8.0,t=1.0,
        lamb=12.0,dl=3.0,lambz=10.0,b=0.0,phi=0.0):
    """Hopping for twisted bilayer graphene"""
    ti = obj2callable(ti) # convert to callable
    cutoff2 = cutoff**2 # cutoff in distance
    def fun(r1,r2):
        rr = (r1-r2) # distance
        rm  = (r1+r2)/2. # average location
        rr = rr.dot(rr) # distance
        if rr>cutoff2: return 0.0 # too far
        if rr<0.001: return 0.0 # same atom
        dx = r1[0]-r2[0]
        dy = r1[1]-r2[1]
        dz = r1[2]-r2[2]
        r = np.sqrt(rr)
        if (r-1.0)<-0.1: raise
        out = -t*(dx*dx + dy*dy)/rr*np.exp(-lamb*(r-1.0))*np.exp(-lambz*dz*dz)
        # interlayer hopping
        out += -ti(rm)*(dz*dz)/rr*np.exp(-lambi*(r-dl))
        #### fix for magnetic field
  #      cphi = np.cos(phi*np.pi)
  #      sphi = np.sin(phi*np.pi)
  #      r = (r1+r2)/2.
  #      dr = r1-r2
  #      p = 2*r[2]*(dr[0]*sphi - dr[1]*cphi)
  #      out *= np.exp(1j*b*p)
        #####
        return out
    return fun


def real_graphene():
    return twisted(ti=0.15,lamb=0.5)



def twisted_matrix(cutoff=5.0,ti=0.3,lambi=8.0,mint=1e-5,
        t=1.0,lamb=12.0,dl=3.0,lambz=10.0,**kwargs):
  """Function capable of returning the hopping matrix
  for twisted bilayer graphene"""
#  if use_fortran:
#    from . import specialhoppingf90
#    def funhop(r1,r2):
#      """Function that returns a hopping matrix"""
#      nr = len(r1) # 
#      nmax = len(r1)*int(10*cutoff**2) # maximum number of hoppings
#      (ii,jj,ts,nout) = specialhoppingf90.twistedhopping(r1,r2,nmax,
#                                  cutoff,ti,lamb,lambi,lambz,mint,dl)
#      if nout>nmax: raise # sanity check
#      ts = ts[0:nout]
#      ii = ii[0:nout]
#      jj = jj[0:nout]
#      out = csc_matrix((ts,(ii-1,jj-1)),shape=(nr,nr),dtype=np.complex_) # matrix
#      return out
#    return funhop # return function
  if True:
      if callable(ti): # workaround for callable interlayer hopping
          tij = twisted(cutoff=cutoff,ti=ti,t=t,
                          lambi=lambi,lamb=lamb,dl=dl,lambz=lambz)
          return entry2matrix(tij)
      else: # conventional JIT function
          return twisted_matrix_python(cutoff=cutoff,ti=ti,mint=mint,
                                  t=t,
                                  lambi=lambi,lamb=lamb,lambz=lambz,
                                  dl=dl,**kwargs)


def multilayer(ti=0.3,dz=3.0):
    """Return hopping for a multilayer"""
    def fhop(ri,rj):
      """Function to compute the hopping"""
      dr = ri-rj ; dr2 = dr.dot(dr) # distance
      if abs(1.0-dr2)<0.01: return 1.0 # first neighbors
      # interlayer hopping (distance between the layers is 3)
      if abs(dz**2-dr2)<0.01 and abs(dz-abs(dr[2]))<0.01: return ti
      return 0.0 # else
    return fhop

def entry2matrix(f):
    def fout(rs1,rs2):
        return np.array([[f(r1,r2) for r1 in rs1] for r2 in rs2]).T
    return fout


def phase_C3_matrix(*args,**kwargs):
    f = phase_C3(*args,**kwargs)
    return entry2matrix(f) # return the matrix


def phase_C3(g,phi=0.5,t=1.0,d=1.0):
    """Create a function that computes hoppings that alternate
    between +\phi and -\phi every 60 degrees"""
    if len(g.r)==1:
      g = g.get_supercell(2) # create a supercell
    ds = g.get_connections()
    i = 0 # first site
    j = ds[i][0] # connected site
    dr = g.r[i] - g.r[j] # distance between sites
    z = dr[0] + 1j*dr[1] # comple vector
    z0 = np.exp(1j*np.pi*2./3.) # rotation
    zs = [z,z*z0,z*z0**2] # the three vectors
    def fun(r1,r2):
        """Function to compute hoppings"""
        dr = r1-r2
        if (d-0.01)<dr.dot(dr)<(d+0.01): # first neighbors
            zi = dr[0]+1j*dr[1]
            for zj in zs: # one of the three directions
                dd = np.abs(zi/zj-1.0)
                if dd<1e-2: 
                    return t*np.exp(1j*phi*np.pi)
            else: return np.exp(-1j*phi*np.pi)
        return 0.0
    return fun


def neighbor_hopping_matrix(g,vs):
    """Return a hopping matrix for the N first neighbors"""
    ds = g.neighbor_distances(n=len(vs)) # get the different distances
    ds = ds[0:len(vs)] # take only these
    return distance_hopping_matrix(vs,ds)




def distance_hopping_matrix(vs,ds):
    """Return a hopping that to the 1-th neighbor is vs"""
    vs = np.array(vs)
    ds = np.array(ds)
    def mgenerator(r1,r2):
        r1 = np.array(r1)
        r2 = np.array(r2)
        n = len(r1)
        out = np.zeros((n,n),dtype=np.complex_) # output
        return distance_hopping_matrix_jit(r1,r2,vs,ds*ds,out) 
    return mgenerator

@jit(nopython=True)
def distance_hopping_matrix_jit(r1,r2,vs,ds2,out):
    """Return a hopping that to the 1-th neighbor is vs"""
    n = len(r1) # number of sites
    nn = len(ds2) # number of neighbors
    for i in range(n):
      for j in range(n):
          dr = r1[i] - r2[j] # difference
          dr2 = dr[0]*dr[0] + dr[1]*dr[1] + dr[2]*dr[2]
          for k in range(nn):
              if np.abs(ds2[k]-dr2)<1e-4: out[i,j] = vs[k]
    return out



def strained_hopping(g,t=1.0,dt=0.0,f=None,**kwargs):
    """Return first neighbor hoppings with strain"""
    if f is None: # no function provided
        from . import potentials
        f = potentials.commensurate_potential(g,average=t,amplitude=dt,
                **kwargs)
    def fout(r1,r2):
        dr = r1-r2
        dr2 = dr.dot(dr)
        if .9<dr2<1.1: return f((r1+r2)/2.)
        else: return 0.0
    return fout


def strained_hopping_matrix(*args,**kwargs):
    f = strained_hopping(*args,**kwargs)
    return entry2matrix(f) # return the matrix


from .algebra import isnumber

class HoppingGenerator():
    """Class for a Hopping generator"""
    def __init__(self,m):
        if type(m)==HoppingGenerator: self.f = m.f # redefine
        elif callable(m): self.f = m # define callable function
        else: raise
    def __call__(self,rs1,rs2):
        """Call method"""
        return self.f(rs1,rs2)
    def __mul__(self,m):
        if isnumber(m): # number input 
            fout = lambda rs1,rs2: np.array(m*np.array(self(rs1,rs2)))
        else: # anything else
            m2 = HoppingGenerator(m) # transform to HoppingGenerator
            fout = lambda rs1,rs2: np.array(self(rs1,rs2))*np.array(m2(rs1,rs2))
        return HoppingGenerator(fout) # return new object
    def __rmul__(self,m): return self*m # commutative 
    def __radd__(self,m): return self + m # commutative 
    def __add__(self,m):  
        m2 = HoppingGenerator(m) # transform to HoppingGenerator
        fout = lambda rs1,rs2: np.array(self(rs1,rs2)) + np.array(m2(rs1,rs2))
        return HoppingGenerator(fout) # return new object
    def __neg__(self): return (-1)*self # minus
    def __sub__(self,a): return self + (-1)*a # minus
    def copy(self):
        from copy import deepcopy
        return deepcopy(self)
    def apply(self,f):
        """Given a certain function of the position, 
        apply it to all the hoppings"""
        def fm(rs1,rs2): # define a new function
            m = self.f(rs1,rs2) # get the matrix
            p = np.array([[f((r1+r2)/2.) for r1 in rs1] for r2 in rs2])
            return m*p # return product
        return HoppingGenerator(fm) # new generator




def twisted_matrix_python(cutoff=10,**kwargs):
  """Function returning the hopping of a twisted matrix"""
  def tij(rs1,rs2): # function to return
      nr = len(rs1) # length
      nmax = nr*int(10*cutoff**2) # maximum number of hoppings
      data = np.zeros(nmax,dtype=np.complex_) # data
      ii = np.zeros(nmax,dtype=int) # index
      jj = np.zeros(nmax,dtype=int) # index
      ii,jj,data,nk = twisted_matrix_jit(np.array(rs1),np.array(rs2),
                         ii,jj,data,cutoff=cutoff,**kwargs) # call function
      if nk>nmax: raise # sanity check
      ii = ii[0:nk] # only nonzero
      jj = jj[0:nk] # only nonzero
      data = data[0:nk] # only nonzero
      out = csc_matrix((data,(ii,jj)),shape=(nr,nr),dtype=np.complex_) # matrix
      out.eliminate_zeros()
      return out
  return tij
    


@jit(nopython=True)
def twisted_matrix_jit(rs1,rs2,ii,jj,data,cutoff=5.0,ti=0.3,lambi=8.0,
        mint = 1e-5,t=1.0,
        lamb=12.0,dl=3.0,lambz=10.0):
  """Hopping for twisted bilayer graphene, returning the indexes of a 
  sparse matrix"""
  cutoff2 = cutoff**2 # cutoff in distance
  ik = 0 # counter
  for i1 in range(len(rs1)):
    for i2 in range(len(rs2)):
      r1 = rs1[i1]
      r2 = rs2[i2]
      dr = r1-r2 # distance
      rr = dr[0]**2 + dr[1]**2 + dr[2]**2 # distance
      if rr>cutoff2: continue  # too far
      if rr<0.001: continue # same atom
      dx = r1[0]-r2[0]
      dy = r1[1]-r2[1]
      dz = r1[2]-r2[2]
      r = np.sqrt(rr)
  #    if r2>100.0: return 0.0 # too far
      if (r-1.0)<-0.1:
        raise
      out = -t*(dx*dx + dy*dy)/rr*np.exp(-lamb*(r-1.0))*np.exp(-lambz*dz*dz)
      out += -ti*(dz*dz)/rr*np.exp(-lambi*(r-dl))
      #### fix for magnetic field
#      cphi = np.cos(phi*np.pi)
#      sphi = np.sin(phi*np.pi)
#      r = (r1+r2)/2.
#      dr = r1-r2
#      p = 2*r[2]*(dr[0]*sphi - dr[1]*cphi)
#      out *= np.exp(1j*b*p)
      #####
      if np.abs(out)>mint:
        ii[ik] = i1 # store
        jj[ik] = i2 # store
        data[ik] = out # store
        ik += 1 # increase counter
  return ii,jj,data,ik





def NNG(g,ts):
    """Hopping generator for nearest neighbor hoppings
    - g: geometry
    - ts: hoppings for each neighbor
    """
    fm = neighbor_hopping_matrix(g,ts) # get the function
    return HoppingGenerator(fm) # return a generator 

def ILG(g,ti,**kwargs):
    """Generator for an interlayer hopping using an exponential
    parametrization
    - g: geometry
    - ti: interlayer hopping
    - **kwargs: optional arguments for twisted hopping
    """
    from .potentials import Potential
    from . import algebra
    if callable(ti): ti = Potential(ti) # transform to potential
    elif algebra.isnumber(ti): pass
    else: raise # not implemented
    fm = twisted_matrix(t=0.,ti=-1*ti,**kwargs) # interlayer hopping generator
    # return a generator
    return HoppingGenerator(lambda *args: algebra.todense(fm(*args))) 



import numpy as np

from ..geometry import Geometry

def sum_geometries(g1,g2):
    """Sum two geometries"""
    if type(g2)==Geometry:
        if g1.dimensionality!=g2.dimensionality: raise
        g = g1.copy()
        g.r = np.concatenate([g1.r,g2.r])
        g.r2xyz()
        if g.has_sublattice:
            g.sublattice = np.concatenate([g1.sublattice,g2.sublattice])
        if g.atoms_have_names:
            g.atoms_names = np.concatenate([g1.atoms_names,g2.atoms_names])
        if g1.primal_geometry is not None and g2.primal_geometry is not None:
            print("Primal geometry not implemented in __add__")
            raise
        return g
    elif type(g2)==np.ndarray: # array input
        g = g1.copy() # copy geometry
        g.r = g.r + g2 # shift all the positions
        g.r2xyz()
        return g
    else:
        print(type(g2))
        raise




# summy class for the kappa method

import numpy as np

class DualLocalProbe():
    def __init__(self,LP):
        self.SC_LP = LP.copy() 
        self.normal_LP = LP.copy()
        self.normal_LP.remove_pairing()
    def set_i(self,i):
        self.SC_LP.i = i
        self.normal_LP.i = i
    def get_kappa(self,**kwargs):
        from ..transporttk.kappa import get_kappa_ratio
        return get_kappa_ratio(self,**kwargs)




def get_kappa(self,T=1e-2,write=True,nsuper=1,**kwargs):
    from ..transporttk.localprobe import LocalProbe
    lp = LocalProbe(self,T=T,**kwargs) # local probe object
    lp.reuse_gf = True # reuse the Green's function
    # now we will overwrite a few objects
    # this is not very elegant, but it works
    g = lp.H.geometry.get_supercell(nsuper) # supercell geometry
    Hc = lp.H.copy() # copy the original object
    # for the selfenernergy, the intracell is picked from lp.H
    lp.H = self.H.get_supercell(nsuper) # overwrite Hamiltonian (for the intra)
    # the Green's function is now directly computed for the supercell
    lp.H.get_gf = lambda **kwargs: Hc.get_gf(nsuper=nsuper,**kwargs)
    dlp = DualLocalProbe(lp) # create a dual local probe
    # now that the methods are overwritten, lets compute
    ks = [] # conductances
    for i in range(len(g.r)): # loop over positions
        dlp.set_i(i) # update position
        ks.append(dlp.get_kappa(**kwargs)) # compute this site
    if write:
        np.savetxt("KAPPAS.OUT",np.array([g.r[:,0],g.r[:,1],np.array(ks)]).T)
    return g.r[:,0],g.r[:,1],np.array(ks)










from __future__ import print_function
import numpy as np
from scipy.sparse import coo_matrix, bmat, csc_matrix
import scipy.sparse as sp
from scipy.sparse import issparse
from . import algebra



def get_eh_sector_odd_even(m,i=0,j=0):
    """ Return the electron hole sector of a matrix,
    assumming that the matrix is in full nambu form""" 
    if i>1 or j>1: return NotImplemented # meaningless
    m = nambu2block(m) # reorder the matrix
    n = m.shape[0]//2 # number of orbitals
    if i==0 and j==0: return m[0:n,0:n] 
    elif i==1 and j==0: return m[n:2*n,0:n] 
    elif i==0 and j==1: return m[0:n,n:2*n] 
    elif i==1 and j==1: return m[n:2*n,n:2*n] 
    else: raise


get_eh_sector = get_eh_sector_odd_even

def get_electron_sector(m):
    """Return the electron sector"""
    return get_eh_sector_odd_even(m,i=0,j=0)


def get_anomalous_sector(m):
    """Return the anomalous electron-hole sector"""
    return get_eh_sector_odd_even(m,i=0,j=1)


def get_second_anomalous_sector(m):
    """Return the anomalous electron-hole sector"""
    return get_eh_sector_odd_even(m,i=1,j=0)


def get_nambu_tauz(m,has_eh=False):
  """Return the nambu matrix tauz for electron-hole"""
  raise # this function is not consistent with the Nambu notation (see below)
  n = m.shape[0] # number of sites 
  if has_eh: n = n//2 # half
  mout = np.matrix(np.zeros((n*2,n*2)),dtype=np.complex_) # define matrix
  for ii in range(n): # loop over index
    mout[2*ii,2*ii] = 1. # assign
    mout[2*ii+1,2*ii+1] = -1. # assign
  return mout # return tauz



def project_electrons(m):
  """Return the nambu matrix tauz for electron-hole"""
  n = m.shape[0] # number of sites 
  mout = m*0.0 # define matrix
  for ii in range(n): # loop over index
    for jj in range(n): # loop over index
      if ii%4<2 and jj%4<2:  mout[ii,jj] = m[ii,jj] # assign
      else: continue
  return mout # return tauz



def project_holes(m):
  """Return the nambu matrix tauz for electron-hole"""
  n = m.shape[0] # number of sites 
  mout = m*0.0 # define matrix
  for ii in range(n): # loop over index
    for jj in range(n): # loop over index
      if ii%4>1 and jj%4>1:  mout[ii,jj] = m[ii,jj] # assign
      else: continue
  return mout # return tauz




def eh_operator(m):
  """Return the electron hole symmetry operator, as a function"""
  n = m.shape[0]//4 # number of sites 
  from .hamiltonians import sy
  msy = [[None for ri in range(n)] for j in range(n)]
  for i in range(n): msy[i][i] = sy # add sy
  msy = bmat(msy) # sy matrix in the electron subspace
  out = [[None,1j*msy],[-1j*msy,None]]
  out = bmat(out) # sparse matrix
  out = reorder(out) # reshufle the matrix
  def ehop(inm):
    """Function that applies electron-hole symmetry to a matrix"""
    return out@np.conjugate(inm)@out # return matrix
  return ehop # return operator
    

def enforce_multihopping_eh_symmetry(MH):
    """Enforce electron-hole symmetry in a multihopping object"""
    raise
    from .multihopping import MultiHopping
    dd = MH.get_dict() # get the dictionary
    out = dict() # dictionary
    feh = eh_operator(dd[(0,0,0)]) # eh operator
    for key in dd: # loop
        m = dd[key] + feh(dd[tuple(-np.array(key))])
        m = m/2. # normalize
        out[key] = m
    return MultiHopping(out) # return object


def enforce_eh_symmetry(cc,cdcd):
    """Enforce electron-hole symmetry, given the two dictionaries
    for the expectation values the anomalous expectation values"""
    outcc = dict() # dictionary
    outcdcd= dict() # dictionary
    for key in cc: # loop over keys
        keym = tuple(-np.array(key)) # the opposite direction
        m = cc[key] # get this correlator (the cc one)
        m2 = cdcd[keym] # get the other matrix
        md = algebra.dagger(m) # get the dagger
        m2d = algebra.dagger(m2) # get the dagger of the matrix
 #       print("MATRICES")
#        print(np.round(m,2))
#        print(np.round(m2,2))
 #       print("MATRICES")
        # this enforces even superconductivity
        outcc[key] = (m + m2d)/2.
        outcdcd[keym] = (m2+md)/2.
    return (outcc,outcdcd) # return the dictionaries





#    raise # this is not ok
#    if type(m)==dict: 
#        from .multihopping import MultiHopping
#        m = MultiHopping(m)
#        return enforce_multihopping_eh_symmetry(m).get_dict()
#    else: raise




def non_unitary_pairing(v):
  """Calculate the vector that defines the non-unitary pairing,
  the input is the three components of the pairing matrix"""
  vc = [np.conjugate(v[i]) for i in range(3)]
  ux = v[1]*vc[2] - v[2]*vc[1]
  uy = v[2]*vc[0] - v[0]*vc[2]
  uz = v[0]*vc[1] - v[1]*vc[0]
  return (1j*np.array([ux,uy,uz])).real



# from now on, the spinors will be
# Psi =
# Psi_up
# Psi_dn
# Psi^dag_dn
# -Psi^dag_up


spair = csc_matrix([[0.,0.,1.,0.],[0.,0.,0.,1.],[0.,0.,0.,0.],[0.,0.,0.,0.]])
proje = csc_matrix([[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,0.,0.],[0.,0.,0.,0.]])
projh = csc_matrix([[0.,0.,0.,0.],[0.,0.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]])
#pzpair = csc_matrix([[0.,0.,0.,1.],[0.,0.,-1.,0.],[0.,0.,0.,0.],[0.,0.,0.,0.]])
#pxpair = csc_matrix([[0.,0.,1.,0.],[0.,0.,0.,-1.],[0.,0.,0.,0.],[-0.,0.,0.,0.]])
#pypair = csc_matrix([[0.,0.,-1j,0.],[0.,0.,0.,1j],[0.,0,0.,0.],[0,0.,0.,0.]])
# Definition using Manfred's notes
# This are the operators that give the d-vectors
# Minus sign due to the nambu spinor!
deltauu = csc_matrix([[0.,0.,0.,0.],[0.,0.,0.,0.],[0.,0.,0.,0.],[-1.,0.,0.,0.]])
# and the rest are simple
deltadd = csc_matrix([[0.,0.,0.,0.],[0.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,0.,0.]])
deltauu = deltauu.H
deltadd = deltadd.H
deltax = (deltadd - deltauu)/2.
deltay = (deltadd + deltauu)/2j
# this one is tricky, we only want the antisymmetric part
deltaz = csc_matrix([[0.,0.,0.,0.],[0.,0.,0.,0.],[1.,0.,0.,0.],[0.,-1.,0.,0.]])/2.

deltaz = deltaz.H



# redefine the functions according to the previous notation

def time_reversal(m):
  """Do the spinful time reversal of this matrix"""
  from .hamiltonians import sy
  n = m.shape[0]//2 # number of spinful blocks
  msy = [[None for ri in range(n)] for j in range(n)]
  for i in range(n): msy[i][i] = sy # add sy
  msy = bmat(msy) # as block matrix
  return msy*np.conjugate(m)*msy # return time reversal




def build_eh(hin,coupling=None,is_sparse=True):
  if coupling is not None:
    c12 = coupling
    c21 = np.conjugate(coupling).T
  else:
    c12 = None
    c21 = None
  return build_nambu_matrix(hin,is_sparse=is_sparse,c12=c12,c21=c21)


def build_nambu_matrix(hin,c12=None,c21=None,is_sparse=True):
    n = hin.shape[0]  # dimension of input
    bout = [[None,None],[None,None]] # initialize None matrix
    bout[0][0] = csc_matrix(hin) # electron part
    bout[1][1] = -time_reversal(csc_matrix(hin)) # hole part
    if c12 is not None:
      bout[0][1] = csc_matrix(c12) # pairing part
    if c21 is not None:
      bout[1][0] = csc_matrix(c21) # pairing part
    bout = bmat(bout) # return matrix
    out = reorder(bout) # reorder matrix
    if issparse(hin): return out
    else: return out.todense()






def add_swave(delta=0.0,is_sparse=False,rs=None):
  """ Adds swave pairing """
  if rs is None: raise # raise error to signal that this is temporal
  n = len(rs) # number of sites
  if callable(delta): # delta is a function
    datar = [delta(ri) for ri in rs] # generate data for the different positions
    data = []
    # the coupling involves up and down
    for dr in datar: # loop over positions
      data.append(dr)  # up e dn h
      data.append(dr)  # dne up h
    iis = range(n*2) # indexes
    coupling = csc_matrix((data,(iis,iis)),dtype=np.complex_)  # generate matrix
  else:
    coupling = sp.identity(n*2)*delta # delta matrix
  zero = coupling*0.
  return build_eh(zero,coupling=coupling,is_sparse=is_sparse) # return matrix


def add_swave_to_hamiltonian(self,delta,**kwargs):
    """Add the swave coupling to the Hamiltonian"""
    from .operators import isnumber
#    if isnumber(delta):
#        if delta==0.0: return 
    # spinless Hamiltonian
    if self.check_mode("spinless") or self.check_mode("spinless_nambu"): 
        from .sctk import spinless
        spinless.add_swave_to_hamiltonian(self,delta)
        # spinful Hamiltonian
    elif self.check_mode("spinful") or self.check_mode("spinful_nambu"): 
      self.turn_nambu() # add electron hole
      self.intra = self.intra + add_swave(delta=delta,rs=self.geometry.r,is_sparse=self.is_sparse)
    else: raise









def add_pxipy(delta=0.0,is_sparse=False,r1=None,r2=None):
  """Add px x + py y pairing"""
  def deltafun(r1i,r2i):
    """Function to calculate the pairing"""
    dr = r2i-r1i
    dr2 = dr.dot(dr)
    if 0.9<dr2<1.1: # first neighbor
      dr = delta*dr/np.sqrt(dr2) # unit vector
#      dr = np.cross(dr,np.array([0.,0.,1.]))
      dr = [dr[0],1j*dr[1],0.]      
#      dr = [0.,0.,dr[0]+1j*dr[1]]
#      return dr
      return dvector2deltas(dr) # return delta
    else: return [0.,0.,0.] # zero vector
  return add_pairing(deltas=deltafun,r1=r1,r2=r2)


add_pwave = add_pxipy


# functions for d-vector computations
from .sctk.dvector import dvector2deltas 
from .sctk.dvector import delta2dvector 



def add_pairing(deltas=[[0.,0],[0.,0.]],is_sparse=True,r1=[],r2=[]):
  """ Adds a general pairing in real space"""
  def get_pmatrix(r1i,r2j): # return the different pairings
    if callable(deltas): dv = deltas(r1i,r2j) # get the components
    else: dv = deltas
    duu = dv[0,1] # delta up up
    ddd = dv[1,0] # delta dn dn
    dud = dv[0,0] # delta up dn
    ddu = dv[1,1] # delta up dn
    # be aware of the minus signs coming from the definition of the
    # Nambu spinor!!!!!!!!!!!!!!!!!
    # c_up d_dn d^\dagger_dn -d^\dagger_up
    D = csc_matrix([[dud,duu],[ddd,ddu]]) # SC matrix
#    D = bmat([[None,D],[-D.H,None]]) # the minus sign comes from triplet
    return D
  # superconducting coupling
  n = len(r1)  # number of sites
  pout = [[None for i in range(n)] for j in range(n)] # initialize None matrix
  # zeros in the diagonal
#  for i in range(n): bout[i][i] = csc_matrix(np.zeros((4,4),dtype=np.complex_))
  for i in range(n): # loop over sites
    for j in range(n): # loop over sites
      pout[i][j] = get_pmatrix(r1[i],r2[j]) # get this pairing
  diag = sp.identity(2*len(r1))*0. # zero matrix
#  diag = csc_matrix(np.zeros((2*len(r1),2*len(r1)),dtype=np.complex_)) # diag
  pout = bmat(pout) # convert to block matrix
#  mout = [[diag,pout],[np.conjugate(pout),diag]] # output matrix
  mout = [[diag,pout],[None,diag]] # output matrix
#  mout = [[diag,pout],[pout,diag]] # output matrix
  mout = bmat(mout) # return full matrix
  mout = reorder(mout) # reorder the entries
  return mout




from .sctk.reorder import block2nambu
from .sctk.reorder import block2nambu_matrix
from .sctk.reorder import nambu2block

reorder = block2nambu


from .sctk.extract import extract_pairing
from .sctk.extract import extract_singlet_pairing
from .sctk.extract import extract_triplet_pairing
from .sctk.pairing import pairing_generator
from .sctk.fastdeltaud import hopping2deltaud




def add_pairing_to_hamiltonian(self,**kwargs):
    """ Add a general pairing matrix to a Hamiltonian"""
#    self.get_eh_sector = get_eh_sector_odd_even # assign function
    df = pairing_generator(self,**kwargs) # function that outputs a 2x2 matrix
    self.turn_nambu() # add electron hole terms
    r = self.geometry.r # positions 
    m = add_pairing(df,r1=r,r2=r) # intra cell
    self.intra = self.intra + m + algebra.dagger(m)
    if self.dimensionality>0:
      if not self.is_multicell: # for multicell hamiltonians
        self.turn_multicell()
#        print("Converting Hamiltonian to multicell")
      from .multicell import Hopping
      for d in self.geometry.neighbor_directions(): # loop over directions
        # this is a workaround to be able to do triplets
        # do it for +k and -k
        if d.dot(d)<0.0001: continue # skip onsite
        r2 = self.geometry.replicas(d=d) # positions
        m = add_pairing(df,r1=r,r2=r2) # new matrix
#        m2 = add_pairing(df,r1=r,r2=r2) # new matrix, the other way
        m2 = m
        if np.max(np.abs(m))>0.0001: # non zero
            self.hopping.append(Hopping(d=d,m=m)) # add pairing
        if np.max(np.abs(m2))>0.0001: # non zero
            self.hopping.append(Hopping(d=-np.array(d),m=m2.H)) # add pairing
      from .multicell import collect_hopping
      self.hopping = collect_hopping(self)


iden = np.matrix([[1.,0.],[0.,1.]],dtype=np.complex_)
taux = np.matrix([[0.,1.],[1.,0.]],dtype=np.complex_)
tauz = np.matrix([[1.,0.],[0.,-1.]],dtype=np.complex_)




def extract_euphdn(m):
    """Extract electron up hole down sector"""
    n = m.shape[0]//2 # half the dimension
    out = np.matrix(np.zeros((n,n),dtype=np.complex_))
    for i in range(n):
      for j in range(n):
        out[i,j] = m[2*i,2*j]
    return out



def superconductivity_type(h):
    """Check the sype of superconductivity"""
    h.check() # check that everything is ok
    hk = h.get_hk_gen() # get geenrator
    k = np.random.random(3) # random kpoint
    h1 = hk(k) # Hamiltonian at +k
    h2 = hk(-k) # Hamiltonian at -k
    (t,t,m1) = extract_pairing(h1)
    (t,t,m2) = extract_pairing(h2)
#    m3 = time_reversal(m1) # time reversal in spin space
    if np.max(np.abs(m1-m2))<1e-5:
        print("Even pairing")
    elif np.max(np.abs(m1+m2))<1e-5:
        print("Odd pairing")


def get_nambu2signless(m):
    """Get the matrix that transforms a SC matrix (with the - sign in the
    spinor), so a basis without - sign"""
    n = m.shape[0]//4 # number of sites
    dd = []
    for i in range(n):
        dd.append(1.)
        dd.append(1.)
        dd.append(1.)
        dd.append(-1.)
    from scipy.sparse import diags
    return diags([dd],[0],shape=(4*n,4*n),dtype=np.complex_) # create matrix



def nambu_anomalous_reordering(n):
    """For a matrix of <eu,ed>, return a matrix that reorders
    to the Nambu basis"""
    m = np.zeros((2*n,2*n)) # initialize as zero matrix
    for i in range(n): # loop over sites
        m[2*i,2*i+1] = 1.0 # reorder
        m[2*i+1,2*i] = 1.0 # reorder
    return m # return the matrix




def identify_superconductivity(h,tol=1e-5):
    """Given a Hamiltonian, identify the kind of superconductivity"""
    if not h.has_eh: return [] # empty list
    dd = h.get_multihopping()
    if dd.norm()<tol: return [] # nothing
    out = [] # initialize the list
#    out.append("Superconductivity") # is superconducting
    dv = h.get_average_dvector() # get the average d-vector
    if dv[0]>tol: out.append("dx SC")
    if dv[1]>tol: out.append("dy SC")
    if dv[2]>tol: out.append("dz SC")
    (uum,ddm,udm) = dict2absdeltas(dd.get_dict()) # extract absolute value of deltas
    if np.max(uum)>tol: out.append("up-up pairing")
    if np.max(ddm)>tol: out.append("down-down pairing")
    if np.max(udm)>tol: out.append("up-down pairing")
    if np.sqrt(np.sum(h.get_average_dvector(non_unitarity=True)))>tol:
        out.append("Non-unitary superconductivity")
    if h.dimensionality==0: return out
    k = np.random.random(3) # random k-vector
    m1 = h.get_hk_gen()(k) # get the Bloch hamiltonian
    m2 = h.get_hk_gen()(-k) # get the Bloch hamiltonian
    # now check if it has some symmetry
    singletp = np.sum(np.abs(extract_singlet_pairing(m1))) # singlet
    tripletp = np.sum(np.abs(extract_triplet_pairing(m1))) # triplet
    if singletp>tol: out.append("Spin-singlet superconductivity")
    if tripletp>tol: out.append("Spin-triplet superconductivity")
#    print(np.round(d1-d2,2),tol)
    return out



def dict2absdeltas(mf):
    """Given a hopping dictionary, extract the absolute value of
    the three deltas, uu,dd and ud. This function should be used for
    qualitative checks"""
    for key in mf: n = mf[key].shape[0]//4 # number of sites
    uu = np.zeros(n)
    dd = np.zeros(n)
    ud = np.zeros(n)
    for key in mf: # loop over keys
        m = mf[key]
        (uut,ddt,udt) = extract_pairing(m) # extract the three matrices
        uu += np.sum(np.abs(uut),axis=0)
        dd += np.sum(np.abs(ddt),axis=0)
        ud += np.sum(np.abs(udt),axis=0)
    return (uu,dd,ud)






def check_nambu():
    n = 20
    m00 = np.random.random((n,n)) + 1j*np.random.random((n,n))
    m01 = np.random.random((n,n)) + 1j*np.random.random((n,n))
    m10 = np.random.random((n,n)) + 1j*np.random.random((n,n))
    m = build_nambu_matrix(m00,c12=m01,c21=m10) 
    d01 = get_eh_sector(m,i=0,j=1)
    d10 = get_eh_sector(m,i=1,j=0)
    print(np.max(np.abs(d01-m01)))
    print(np.max(np.abs(d10-m10)))




def turn_nambu(self):
  """Turn a Hamiltonian an Nambu Hamiltonian"""
  nambu = build_eh
  if self.check_mode("spinful_nambu"): return # do nothing
  elif self.check_mode("spinless"): self.turn_spinful() # error
  elif self.check_mode("spinful"): pass # error
  else: raise
  def f(m): return nambu(m,is_sparse=self.is_sparse)
  self.modify_hamiltonian_matrices(f) # modify all the matrices
  self.has_eh = True


from .sctk.extract import get_anomalous_hamiltonian
from .sctk.dvector import average_hamiltonian_dvector

def remove_pairing(H):
    has_eh = H.has_eh # check if there is eh
    H.remove_nambu()
    if has_eh: H.setup_nambu_spinor()  # put eh again





# library with specific function for hybrid ribbons
from __future__ import print_function
import numpy as np




def create_hybrid(h1,h2,coupling=1.0):
  """ Creates a hybrid hamiltonian from the two inputs"""
  from copy import deepcopy
  hy = deepcopy(h1) # copy hamiltonian
  # check that thw two hamiltonians are mixable
  if not h1.has_spin == h2.has_spin:  # spinpol
    raise
  if not h1.has_eh == h2.has_eh:  # electron hole
    raise
  if not len(h1.intra) == len(h2.intra):  # same dimension
    print("Wrong dimensions", len(h1.intra),len(h2.intra))
    raise
  if not h1.dimensionality == h2.dimensionality:  # dimension
    raise
  dd = h1.intra.shape[0] # dimension of the hamiltonian
  # we wnt a matrix like
  # ( h1_1   t12  )
  # ( t21   h2_2  )
  # substitute the block h2_2 by the second block in input h2
  for i in range(dd/2,dd):
    for j in range(dd/2,dd):
      hy.intra[i,j] = h2.intra[i,j] # substitute element
      hy.inter[i,j] = h2.inter[i,j] # substitute element
  # and now create t12 by mixing 
  for i in range(dd/2):
    for j in range(dd/2,dd):
      hy.intra[i,j] = coupling*(h1.intra[i,j]+h2.intra[i,j])/2.0 # substitute element
      hy.inter[i,j] = coupling*(h1.inter[i,j]+h2.inter[i,j])/2.0 # substitute element
  for i in range(dd/2,dd):
    for j in range(dd/2):
      hy.intra[i,j] = coupling*(h1.intra[i,j]+h2.intra[i,j])/2.0 # substitute element
      hy.inter[i,j] = coupling*(h1.inter[i,j]+h2.inter[i,j])/2.0 # substitute element
  hy.check()  # check that everything is fine
  return hy  # return the hamiltonian




def half_and_half(h1,h2,fun=None,tlen=0.001,direction=1):
  """Create a Hamiltonian using half one matrix and half of the
  other"""
#  if h1.dimensionality ==0: raise # only for 1d
  if fun is None: # no function provided
    def fun(r):
      if r[direction]<-tlen: return 0.0
      elif r[direction]>tlen: return 1.0
      else: return (r[1]+tlen)/(2*tlen) # interpolate
  hout = h1.copy() # copy Hamiltonians
  d = 1
  if h1.has_spin: d *= 2 # double
  if h1.has_eh: d *= 2 # double
  for i in range(h1.intra.shape[0]):
    for j in range(h1.intra.shape[0]):
      fac = fun((h1.geometry.r[i//d]+h2.geometry.r[j//d])/2) # factor
      hout.intra[i,j] = fac*(h1.intra[i,j]) + (1.-fac)*h2.intra[i,j]
      if h1.is_multicell: # multicell Hamiltonian
        for ii in range(len(h1.hopping)):
          hout.hopping[ii].m[i,j] = fac*(h1.hopping[ii].m[i,j]) + (1.-fac)*h2.hopping[ii].m[i,j]
        
      else:
        if h1.dimensionality==1: # one dimensional
          hout.inter[i,j] = fac*(h1.inter[i,j]) + (1.-fac)*h2.inter[i,j]
        else: raise
  return hout




def heterostructure(h1,h2):
  """Build a heterostructure, assuming that the only difference
  are onsite matrices"""
  h = h1.copy() # copy Hamiltonian
  n = h.intra.shape[0]//2 # retain these elements
  for i in range(n):
    for j in range(n):
      h.intra[i,j] = h2.intra[i,j]
  return h # return Hamiltonian












# routines to transform a pyqula Hamiltonian into a dmrgpy one
from . import algebra
import numpy as np

def generate_fermionic_chain(H):
    """Return a Fermionic object from dmrgpy to solve the Hamiltonian
    exactly"""
    from dmrgpy import fermionchain
    if H.dimensionality!=0:
        print("Only implemented for 0d dimensional Hamiltonians")
    n = H.intra.shape[0] # number of total orbitals
    if H.has_spin:
        fc = fermionchain.Spinful_Fermionic_Chain(n//2) # generate chain
    else:
        fc = fermionchain.Fermionic_Chain(n) # generate chain
    hp = 0. # initialize
    m = algebra.todense(H.intra) # get the dense matrix
    for i in range(n):
        for j in range(n):
            tij = m[i,j] # hopping
            if np.abs(tij)>1e-5:
                hp = hp + tij*fc.Cdag[i]*fc.C[j]
    fc.set_hamiltonian(hp) # set up the Hamiltonian
    return fc # return the fermionic chain object





import numpy as np

from .. import algebra

# define an embedded object based on a Hamiltonian

class Embedded_Hamiltonian():
    def __init__(self,H,delta=1e-4,selfenergy=None):
        self.H = H.copy() # copy Hamiltonian
        self.selfenergy = object2selfenergy(selfenergy,H)
        self.delta = delta
    def get_density_matrix(self,**kwargs): 
        return get_dm(self,delta=self.delta,**kwargs)
    def get_gf(self,**kwargs):
        # Green's function of the Hamiltonian
        gf0 = self.H.get_gf(**kwargs) 
        selfe = self.selfenergy(**kwargs) # store selfenergy
        gf = algebra.inv(algebra.inv(gf0) - selfe) # full Green's function
        return gf # return full Green's function
    def set_multihopping(self,*args): 
        self.H.set_multihopping(*args)
    def get_mean_field_hamiltonian(self,**kwargs):
        from ..selfconsistency.embedding import hubbard_mf
        return hubbard_mf(self,**kwargs) # return Hubbard mean-field
    def copy(self):
        from copy import deepcopy
        return deepcopy(self)
    def get_ldos(self,**kwargs): 
        A = get_A(self,**kwargs) # spectral function
        r = [A[i,i].real for i in range(A.shape[0])]
        r = self.H.full2profile(r) # resum components
        self.H.geometry.write_profile(r,name="LDOS.OUT")
        return H.geometry.r,r # return LDOS
    def get_kdos(self,**kwargs): return get_kdos(self,**kwargs)



def object2selfenergy(self,H,delta=1e-4,**kwargs):
    if self is None: # no selfenergy provided 
        return lambda **kwargs: 0.
    elif algebra.ismatrix(self): # matrix provided
      if H.intra.shape[0]==self.shape[0]: 
        def f(delta=delta,**kwargs):
            if delta>0.: return self
            else: return np.conjugate(self)
        return f
      else: raise # unrecognized
    elif callable(self): return self # assume that it returns a matrix
    else: 
        print("Selfenergy is not compatible with Hamiltonian")
        raise


def embed_hamiltonian(self,**kwargs):
   EB = Embedded_Hamiltonian(self,**kwargs)
   return EB



def get_dm(self,delta=1e-2,emin=-10.,**kwargs):
    """Get the density matrix"""
    fa = lambda e: self.get_gf(energy=e,delta=delta,**kwargs) # advanced
    fr = lambda e: self.get_gf(energy=e,delta=-delta,**kwargs) # retarded
    from ..integration import complex_contour
    Ra = complex_contour(fa,xmin=emin,xmax=0.,mode="upper") # return the integral
    Rr = complex_contour(fr,xmin=emin,xmax=0.,mode="lower") # return the integral
    return 1j*(Ra-Rr)/(2.*np.pi) # return the density matrix



def get_A(self,delta=1e-3,**kwargs):
    Ra = self.get_gf(delta=delta,**kwargs) # advanced
    Rr = self.get_gf(delta=-delta,**kwargs) # retarded
    return 1j*(Ra-Rr)/(2.*np.pi) # return the spectral fucntion



def get_kdos(self,energies=None,kpath=None,**kwargs):
    """Compute k-resolved DOS"""
    def f(e,k): # function to evaluate
        gf0 = self.H.get_gk_gen(**kwargs)(e=e,k=k) # get Green's function
        selfe = self.selfenergy(energy=e,**kwargs) # selfenergy
        gf = algebra.inv(algebra.inv(gf0) - selfe) # full Green's function
        return -np.trace(gf).imag # return full Green's function
    if energies is None: energies = np.linspace(-1.0,1.0,100)
    kpath = self.H.geometry.get_kpath(kpath=kpath) # get the kpath
    fo = open("KDOS.OUT","w")
    for ik in range(len(kpath)):
        print("Doing",ik)
        for ie in energies:
            d = f(ie,kpath[ik])
            fo.write(str(ik)+" ")
            fo.write(str(ie)+" ")
            fo.write(str(d)+"\n")
    fo.close()







